

"""Defines widgets which may be used as page footers"""

from .. import tag
from . import Widget, FieldArg, FieldArgList, FieldArgTable, FieldArgDict


class SimpleFooter(Widget):
    """Defines a div, with a paragraph of text"""

    error_location = (0,0,0)

    arg_descriptions = {
                        'footer_text':FieldArg("text",'', jsonset=True),
                        'para_class':FieldArg("cssclass",""),
                        'error_class':FieldArg("cssclass", ""),
                        'paradiv_class':FieldArg("cssclass",""),
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
         footer_text: The text to appear in the footer
         paradiv_class: css class of the div holding the footer paragraph
         para_class: css class of the footer text paragraph
         error_class: css class of the error paragraph
        """

        Widget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "div"
        # hidden error
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''
        self[1] = tag.Part(tag_name="div")
        self[1][0] = tag.Part(tag_name="p")

    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the paragraph"
        self[0].set_class_style(self.wf.error_class)
        if self.error_status:
            del self[0].attribs["style"]
        self[1].set_class_style(self.wf.paradiv_class)
        self[1][0].set_class_style(self.wf.para_class)
        if self.wf.footer_text:
            self[1][0][0] = self.wf.footer_text
        # set an id in the footer_text paragraph
        # any label:value added to self.jlabels will be set in a javascript fieldvalues attribute for the widget
        self.jlabels['textident'] = self[1][0].insert_id()


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """<div> <!-- with widget id and class widget_class -->
  <div> <!-- normally hidden div, with class error_class -->
    <p>       <!-- Any error text appears here --> </p>
  </div>
  <div>  <!-- with class paradiv_class -->
    <p>  <!-- with class para_class -->
    <!-- footer_text appears in this paragraph -->
    </p>
  </div>
</div>
"""






"""Contains widgets for inputting text"""

from .. import tag
from . import Widget, ClosedWidget, FieldArg, FieldArgList, FieldArgTable, FieldArgDict


class TextInput1(ClosedWidget):
    """Defines an input field only (no form, label or submit button)
       does not display errors"""

    # This class does not display any error messages
    display_errors = False

    # js_validators is a class attribute, True if javascript validation is enabled
    js_validators=True

    arg_descriptions = {
                        'input_accepted_class':FieldArg("cssclass", ''),
                        'input_errored_class':FieldArg("cssclass", ''),
                        'set_input_accepted':FieldArg("boolean", False, jsonset=True),
                        'set_input_errored':FieldArg("boolean", False, jsonset=True),
                        'input_text':FieldArg("text", '', valdt=True, jsonset=True),
                        'size':FieldArg("text", ''),
                        'maxlength':FieldArg("text", ''),
                        'disabled':FieldArg("boolean", False),
                        'required':FieldArg("boolean", False),
                        'type':FieldArg("text", 'text'),
                        'pattern':FieldArg("text", ''),
                        'title':FieldArg("text", '')
                       }
    def __init__(self, name=None, brief='', **field_args):
        """
        input_accepted_class: A class which can be set on the input field
        input_errored_class: A class which can be set on the input field
        set_input_accepted: If True, input_accepted_class will be set on the input field
        set_errored_accepted: If True, input_errored_class will be set on the input field
        input_text: The default text in the text input field, field name used as name attribute
        size: The number of characters appearing in the text input field
        maxlength: The maximum number of characters accepted in the text input field
        disabled: Set True if the field is to be disabled
        required: Set True to put the 'required' attribute into the input field
        type: the type set on the input field, such as text or email
        pattern: regular expression pattern
        title: helps describe the pattern
        """
        ClosedWidget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "input"
        self.attribs["type"] = "text"


    def _build(self, page, ident_list, environ, call_data, lang):
        "build the input field"
        self.attribs["name"] = self.get_formname('input_text')
        self.attribs["value"] = self.wf.input_text

        if self.wf.set_input_accepted and self.wf.input_accepted_class:
            if self.wf.widget_class:
                self.attribs["class"] = self.wf.widget_class + " " + self.wf.input_accepted_class
            else:
                self.attribs["class"] = self.wf.input_accepted_class
        if self.wf.set_input_errored and self.wf.input_errored_class:
            if self.wf.widget_class:
                self.attribs["class"] = self.wf.widget_class + " " + self.wf.input_errored_class
            else:
                self.attribs["class"] = self.wf.input_errored_class

        # any label:value added to self.jlabels will be set in a javascript fieldvalues attribute for the widget
        self.jlabels['input_accepted_class'] = self.wf.input_accepted_class
        self.jlabels['input_errored_class'] = self.wf.input_errored_class

        if self.wf.size:
            self.attribs["size"] = self.wf.size
        if self.wf.maxlength:
            self.attribs["maxlength"] = self.wf.maxlength
        if self.wf.disabled:
            self.attribs["disabled"] = "disabled"
        if self.wf.required:
            self.attribs["required"] = "required"
        if self.wf.type:
            self.attribs["type"] = self.wf.type
        if self.wf.pattern:
            self.attribs["pattern"] = self.wf.pattern
        if self.wf.title:
            self.attribs["title"] = self.wf.title


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
  <input type='text' />
  <!-- with widget id and class widget_class and input_accepted_class added if
       set_input_accepted is set to True, and input_errored_class added if
       set_input_errored is set to True, (which will remove input_accepted_class) -->
"""


class Password1(ClosedWidget):
    """Defines a password input field only (no form, label or submit button)
       does not display errors"""

    # This class does not display any error messages
    display_errors = False

    # js_validators is a class attribute, True if javascript validation is enabled
    js_validators=True

    arg_descriptions = {
                        'input_accepted_class':FieldArg("cssclass", ''),
                        'input_errored_class':FieldArg("cssclass", ''),
                        'set_input_accepted':FieldArg("boolean", False, jsonset=True),
                        'set_input_errored':FieldArg("boolean", False, jsonset=True),
                        'input_text':FieldArg("text", '', valdt=True, jsonset=True),
                        'size':FieldArg("text", ''),
                        'maxlength':FieldArg("text", ''),
                        'disabled':FieldArg("boolean", False),
                        'required':FieldArg("boolean", False)
                       }
    def __init__(self, name=None, brief='', **field_args):
        """
        input_accepted_class: A class which can be set on the input field
        input_errored_class: A class which can be set on the input field
        set_input_accepted: If True, input_accepted_class will be set on the input field
        set_errored_accepted: If True, input_errored_class will be set on the input field
        input_text: The default text in the text input field, field name used as name attribute
        size: The number of characters appearing in the text input field
        maxlength: The maximum number of characters accepted in the text input field
        disabled: Set True if the field is to be disabled
        required: Set True to put the 'required' attribute into the input field
        """
        ClosedWidget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "input"
        self.attribs["type"] = "password"


    def _build(self, page, ident_list, environ, call_data, lang):
        "build the input field"
        self.attribs["name"] = self.get_formname('input_text')
        self.attribs["value"] = self.wf.input_text

        if self.wf.set_input_accepted and self.wf.input_accepted_class:
            if self.wf.widget_class:
                self.attribs["class"] = self.wf.widget_class + " " + self.wf.input_accepted_class
            else:
                self.attribs["class"] = self.wf.input_accepted_class
        if self.wf.set_input_errored and self.wf.input_errored_class:
            if self.wf.widget_class:
                self.attribs["class"] = self.wf.widget_class + " " + self.wf.input_errored_class
            else:
                self.attribs["class"] = self.wf.input_errored_class

        # any label:value added to self.jlabels will be set in a javascript fieldvalues attribute for the widget
        self.jlabels['input_accepted_class'] = self.wf.input_accepted_class
        self.jlabels['input_errored_class'] = self.wf.input_errored_class

        if self.wf.size:
            self.attribs["size"] = self.wf.size
        if self.wf.maxlength:
            self.attribs["maxlength"] = self.wf.maxlength
        if self.wf.disabled:
            self.attribs["disabled"] = "disabled"
        if self.wf.required:
            self.attribs["required"] = "required"


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
  <input type='password' />
  <!-- with widget id and class widget_class and input_accepted_class added if
       set_input_accepted is set to True, and input_errored_class added if
       set_input_errored is set to True, (which will remove input_accepted_class) -->
"""


class TextInput2(Widget):
    """Defines a div containing a hidden error paragraph
       with a text input field (no form or submit button)
       A label to the left of the text can have a class and text set"""

    # js_validators is a class attribute, True if javascript validation is enabled
    js_validators=True

    error_location = (0,0,0)

    arg_descriptions = {
                        'input_accepted_class':FieldArg("cssclass", ''),
                        'input_errored_class':FieldArg("cssclass", ''),
                        'inputdiv_class':FieldArg("cssclass", ''),
                        'set_input_accepted':FieldArg("boolean", False, jsonset=True),
                        'set_input_errored':FieldArg("boolean", False, jsonset=True),
                        'label':FieldArg("text", 'Your Input:'),
                        'label_class':FieldArg("cssclass", ''),
                        'label_style':FieldArg("cssstyle", ''),
                        'error_class':FieldArg("cssclass", ''),
                        'input_text':FieldArg("text", '', valdt=True, jsonset=True),
                        'input_class':FieldArg("cssclass", ''),
                        'input_disabled_class':FieldArg("cssclass", ''),
                        'size':FieldArg("text", ''),
                        'maxlength':FieldArg("text", ''),
                        'redstar':FieldArg("boolean", False),
                        'redstar_class':FieldArg("cssclass", ''),
                        'redstar_style':FieldArg("cssstyle", ''),
                        'disabled':FieldArg("boolean", False, jsonset=True),
                        'required':FieldArg("boolean", False),
                        'type':FieldArg("text", 'text'),
                        'pattern':FieldArg("text", ''),
                        'title':FieldArg("text", '')
                       }
    def __init__(self, name=None, brief='', **field_args):
        """
        input_accepted_class: A class which can be set on the input field
        input_errored_class: A class which can be set on the input field
        set_input_accepted: If True, input_accepted_class will be set on the input field
        set_errored_accepted: If True, input_errored_class will be set on the input field
        label: The text appearing in a label tag to the left of the input field
        label_class: The css class of the label to the left of the input field
        error_class: The class applied to the hidden paragraph on error.
        inputdiv_class: The css class of the div containing the label and input field
        input_text: The default text in the text input field, field name used as name attribute
        input_class: The css class of the input field
        input_disabled_class: css class of the input field which replaces input_class if disabled is True
        size: The number of characters appearing in the text input area
        maxlength: The maximum number of characters accepted in the text area
        redstar: If True a red asterix is shown by the side of the input field
        disabled: Set True if the field is to be disabled
        required: Set True to put the 'required' flag into the input field
        type: the type set on the input field, such as text or email
        pattern: regular expression pattern
        title: helps describe the pattern
        """
        Widget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "div"
        # error div at 0
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''
        # inputdiv
        self[1] = tag.Part(tag_name="div")
        # the label
        self[1][0] = tag.Part(tag_name="label", hide_if_empty=True)
        # the text input field
        self[1][1] = tag.ClosedPart(tag_name="input", attribs ={"type":"text"})


    def _build(self, page, ident_list, environ, call_data, lang):
        "build the input field"

        self[0].set_class_style(self.wf.error_class)
        if self.error_status:
            del self[0].attribs["style"]

        self[1].set_class_style(self.wf.inputdiv_class)

        self[1][0].set_class_style(self.wf.label_class, self.wf.label_style)
        if self.wf.label:
            self[1][0][0] = self.wf.label

        # set an id in the input field for the 'label for' tag
        # any label:value added to self.jlabels will be set in a javascript fieldvalues attribute for the widget
        self.jlabels['input_id'] = self[1][1].insert_id()
        self[1][0].attribs['for'] = self.jlabels['input_id']

        if self.wf.input_accepted_class:
            self.jlabels['input_accepted_class'] = self.wf.input_accepted_class
        if self.wf.input_errored_class:
            self.jlabels['input_errored_class'] = self.wf.input_errored_class
        if self.wf.input_class:
            self.jlabels['input_class'] = self.wf.input_class
        if self.wf.input_disabled_class:
            self.jlabels['input_disabled_class'] = self.wf.input_disabled_class

        self[1][1].attribs.update({"name":self.get_formname('input_text'), "value":self.wf.input_text})

        if self.wf.size:
            self[1][1].attribs["size"] = self.wf.size
        if self.wf.maxlength:
            self[1][1].attribs["maxlength"] = self.wf.maxlength
        if self.wf.required:
            self[1][1].attribs["required"] = "required"
        if self.wf.type:
            self[1][1].attribs["type"] = self.wf.type
        if self.wf.pattern:
            self[1][1].attribs["pattern"] = self.wf.pattern
        if self.wf.title:
            self[1][1].attribs["title"] = self.wf.title

        if self.wf.input_class:
            input_class = self.wf.input_class
        else:
            input_class = ''

        if self.wf.disabled:
            self[1][1].attribs["disabled"] = "disabled"
            if self.wf.input_disabled_class:
                input_class = self.wf.input_disabled_class
        if self.error_status and self.wf.input_errored_class:
            if input_class:
                input_class = input_class + ' ' + self.wf.input_errored_class
            else:
                input_class = self.wf.input_errored_class
        elif self.wf.set_input_errored and self.wf.input_errored_class:
            if input_class:
                input_class = input_class + ' ' + self.wf.input_errored_class
            else:
                input_class = self.wf.input_errored_class
        elif self.wf.set_input_accepted and self.wf.input_accepted_class:
            if input_class:
                input_class = input_class + ' ' + self.wf.input_accepted_class
            else:
                input_class = self.wf.input_accepted_class
        if input_class:
            self[1][1].attribs["class"] = input_class
        # redstar
        if self.wf.redstar:
            self[1][2] = tag.Part(tag_name="span")
            self[1][2].set_class_style(self.wf.redstar_class, self.wf.redstar_style)
            self[1][2][0] = '*'

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
  <div> <!-- normally hidden div, with class error_class -->
    <p> <!-- Any error text appears here --> </p>
  </div>
  <div>  <!-- class attribute set to inputdiv_class -->
    <label> <!-- with class set to label_class -->
            <!-- content set to label -->
    </label>
    <input type='text' /> <!-- with attributes set appropriately -->
    <span>*</span> <!-- shown if redstar is True -->
  </div>
</div>"""


class Password2(Widget):
    """Defines a div containing a hidden error paragraph
       with a password input field (no form or submit button)
       A label to the left of the text can have a class and text set"""

    # js_validators is a class attribute, True if javascript validation is enabled
    js_validators=True

    error_location = (0,0,0)

    arg_descriptions = {
                        'input_accepted_class':FieldArg("cssclass", ''),
                        'input_errored_class':FieldArg("cssclass", ''),
                        'input_disabled_class':FieldArg("cssclass", ''),
                        'inputdiv_class':FieldArg("cssclass", ''),
                        'set_input_accepted':FieldArg("boolean", False, jsonset=True),
                        'set_input_errored':FieldArg("boolean", False, jsonset=True),
                        'label':FieldArg("text", 'Password:'),
                        'label_class':FieldArg("cssclass", ''),
                        'label_style':FieldArg("cssstyle", ''),
                        'error_class':FieldArg("cssclass", ''),
                        'input_text':FieldArg("text", '', valdt=True, jsonset=True),
                        'input_class':FieldArg("cssclass", ''),
                        'input_style':FieldArg("cssstyle", ''),
                        'size':FieldArg("text", ''),
                        'maxlength':FieldArg("text", ''),
                        'disabled':FieldArg("boolean", False, jsonset=True),
                        'redstar':FieldArg("boolean", False),
                        'redstar_class':FieldArg("cssclass", ''),
                        'redstar_style':FieldArg("cssstyle", ''),
                        'required':FieldArg("boolean", False)
                       }
    def __init__(self, name=None, brief='', **field_args):
        """
        input_accepted_class: A class which can be set on the input field
        input_errored_class: A class which can be set on the input field
        input_disabled_class: css class of the input field which replaces input_class if disabled is True
        inputdiv_class: The css class of the div containing the label and input field
        set_input_accepted: If True, input_accepted_class will be set on the input field
        set_errored_accepted: If True, input_errored_class will be set on the input field
        label: The text appearing in a label tag to the left of the input field
        label_class: The css class of the label to the left of the input field
        error_class: The class applied to the hidden paragraph on error.
        input_text: The default text in the text input field, field name used as name attribute
        input_class: The css class of the input field
        size: The number of characters appearing in the text input area
        maxlength: The maximum number of characters accepted in the text area
        disabled: Set True if the field is to be disabled
        redstar: If True a red asterix is shown by the side of the input field
        required: Set True to put the 'required' flag into the input field
        """
        Widget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "div"
        # error div at 0
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''
        # inputdiv
        self[1] = tag.Part(tag_name="div")
        # the label
        self[1][0] = tag.Part(tag_name="label", hide_if_empty=True)
        # the text input field
        self[1][1] = tag.ClosedPart(tag_name="input", attribs ={"type":"password"})


    def _build(self, page, ident_list, environ, call_data, lang):
        "build the input field"

        self[0].set_class_style(self.wf.error_class)
        if self.error_status:
            del self[0].attribs["style"]

        self[1].set_class_style(self.wf.inputdiv_class)

        self[1][0].set_class_style(self.wf.label_class, self.wf.label_style)
        if self.wf.label:
            self[1][0][0] = self.wf.label

        # set an id in the input field for the 'label for' tag
        self[1][0].attribs['for'] = self[1][1].insert_id()

        # any label:value added to self.jlabels will be set in a javascript fieldvalues attribute for the widget
        if self.wf.input_accepted_class:
            self.jlabels['input_accepted_class'] = self.wf.input_accepted_class
        if self.wf.input_errored_class:
            self.jlabels['input_errored_class'] = self.wf.input_errored_class
        if self.wf.input_class:
            self.jlabels['input_class'] = self.wf.input_class
        if self.wf.input_disabled_class:
            self.jlabels['input_disabled_class'] = self.wf.input_disabled_class

        self[1][1].attribs.update({"name":self.get_formname('input_text'), "value":self.wf.input_text})

        if self.wf.size:
            self[1][1].attribs["size"] = self.wf.size
        if self.wf.maxlength:
            self[1][1].attribs["maxlength"] = self.wf.maxlength
        if self.wf.required:
            self[1][1].attribs["required"] = "required"
        if self.wf.input_style:
            self[1][1].attribs["style"] = self.wf.input_style

        if self.wf.input_class:
            input_class = self.wf.input_class
        else:
            input_class = ''
        if self.wf.disabled:
            self[1][1].attribs["disabled"] = "disabled"
            if self.wf.input_disabled_class:
                input_class = self.wf.input_disabled_class
        if self.error_status and self.wf.input_errored_class:
            if input_class:
                input_class = input_class + ' ' + self.wf.input_errored_class
            else:
                input_class = self.wf.input_errored_class
        elif self.wf.set_input_errored and self.wf.input_errored_class:
            if input_class:
                input_class = input_class + ' ' + self.wf.input_errored_class
            else:
                input_class = self.wf.input_errored_class
        elif self.wf.set_input_accepted and self.wf.input_accepted_class:
            if input_class:
                input_class = input_class + ' ' + self.wf.input_accepted_class
            else:
                input_class = self.wf.input_accepted_class
        if input_class:
            self[1][1].attribs["class"] = input_class
        # redstar
        if self.wf.redstar:
            self[1][2] = tag.Part(tag_name="span")
            self[1][2].set_class_style(self.wf.redstar_class, self.wf.redstar_style)
            self[1][2][0] = '*'


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
  <div> <!-- normally hidden paragraph, with class error_class -->
    <p> <!-- Any error text appears here --> </p>
  </div>
  <div>  <!-- class attribute set to inputdiv_class -->
    <label> <!-- with class set to label_class -->
            <!-- content set to label -->
    </label>
    <input type='password' /> <!-- with attributes set appropriately -->
    <span>*</span> <!-- shown if redstar is True -->
  </div>
</div>"""


class TextInput3(Widget):
    """Defines a div containing a text input field (no form or submit button)
       with a left and right label, does not display errors"""

    # This class does not display any error messages
    display_errors = False

    # js_validators is a class attribute, True if javascript validation is enabled
    js_validators=True

    arg_descriptions = {
                        'input_class':FieldArg("cssclass", ''),
                        'input_style':FieldArg("cssstyle", ''),
                        'input_accepted_class':FieldArg("cssclass", ''),
                        'input_errored_class':FieldArg("cssclass", ''),
                        'set_input_accepted':FieldArg("boolean", False, jsonset=True),
                        'set_input_errored':FieldArg("boolean", False, jsonset=True),
                        'left_label':FieldArg("text", 'Your Input:'),
                        'left_class':FieldArg("cssclass", ''),
                        'left_style':FieldArg("cssstyle", ''),
                        'right_label':FieldArg("text", ''),
                        'right_class':FieldArg("cssclass", ''),
                        'right_style':FieldArg("cssstyle", ''),
                        'input_text':FieldArg("text", '', valdt=True, jsonset=True),
                        'size':FieldArg("text", ''),
                        'maxlength':FieldArg("text", ''),
                        'disabled':FieldArg("boolean", False),
                        'required':FieldArg("boolean", False),
                        'type':FieldArg("text", 'text'),
                        'pattern':FieldArg("text", ''),
                        'title':FieldArg("text", '')
                       }
    def __init__(self, name=None, brief='', **field_args):
        """
        input_class: The css class of the input field
        input_style: The css style of the input field
        input_accepted_class: A class which can be set on the input field
        input_errored_class: A class which can be set on the input field
        set_input_accepted: If True, input_accepted_class will be set on the input field
        set_errored_accepted: If True, input_errored_class will be set on the input field
        left_label: The text displayed to the left of the input field
        left_class: The css class of the label to the left of the input field
        right_label: The text displayed to the right of the input field
        right_class: The css class of the label to the right of the input field
        input_text: The default text in the text input field, field name used as name attribute
        size: The number of characters appearing in the text input field
        maxlength: The maximum number of characters accepted in the text input field
        disabled: Set True if the field is to be disabled
        required: Set True to put the 'required' attribute into the input field
        type: the type set on the input field, such as text or email
        pattern: regular expression pattern
        title: helps describe the pattern
        """
        Widget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "div"
        self[0] = tag.Part(tag_name="label", hide_if_empty=True)
        self[1] = tag.ClosedPart(tag_name="input", attribs = {"type":"text"})
        self[2] = tag.Part(tag_name="label", hide_if_empty=True)


    def _build(self, page, ident_list, environ, call_data, lang):
        "build the input field"
        if self.wf.left_label:
            self[0][0] = self.wf.left_label

        self[0].set_class_style(self.wf.left_class, self.wf.left_style)

        self[1].attribs.update({"name":self.get_formname('input_text'), "value":self.wf.input_text})

        if self.wf.input_class:
            input_class = self.wf.input_class
        else:
            input_class = ''

        if self.wf.set_input_errored and self.wf.input_errored_class:
            if input_class:
                input_class = input_class + ' ' + self.wf.input_errored_class
            else:
                input_class = self.wf.input_errored_class
        elif self.wf.set_input_accepted and self.wf.input_accepted_class:
            if input_class:
                input_class = input_class + ' ' + self.wf.input_accepted_class
            else:
                input_class = self.wf.input_accepted_class

        self[1].set_class_style(input_class, self.wf.input_style)

        if self.wf.size:
            self[1].attribs["size"] = self.wf.size
        if self.wf.maxlength:
            self[1].attribs["maxlength"] = self.wf.maxlength
        if self.wf.disabled:
            self[1].attribs["disabled"] = "disabled"
        if self.wf.required:
            self[1].attribs["required"] = "required"
        if self.wf.type:
            self[1].attribs["type"] = self.wf.type
        if self.wf.pattern:
            self[1].attribs["pattern"] = self.wf.pattern
        if self.wf.title:
            self[1].attribs["title"] = self.wf.title

        if self.wf.right_label:
            self[2][0] = self.wf.right_label

        self[2].set_class_style(self.wf.right_class, self.wf.right_style)

        # set an id in the text input field for the 'label for' tag
        # any label:value added to self.jlabels will be set in a javascript fieldvalues attribute for the widget
        self.jlabels['input_id'] = self[1].insert_id()
        self[0].attribs['for'] = self.jlabels['input_id']
        self[2].attribs['for'] = self.jlabels['input_id']

        if self.wf.input_accepted_class:
            self.jlabels['input_accepted_class'] = self.wf.input_accepted_class
        if self.wf.input_errored_class:
            self.jlabels['input_errored_class'] = self.wf.input_errored_class

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
  <label> <!-- with class set to left_class and content to left_label -->
  </label>
  <input type='text' /> <!-- with attributes set appropriately -->
  <label> <!-- with class set to right_class and content to right_label -->
  </label>
</div>"""


class TextInput4(Widget):
    """Defines a div containing a div with label
       and a div with a text input field (no form or submit button)
       All divs, label and input field can have class set
       No error display"""

    # This class does not display any error messages
    display_errors = False

    # js_validators is a class attribute, True if javascript validation is enabled
    js_validators=True

    arg_descriptions = {
                        'input_accepted_class':FieldArg("cssclass", ''),
                        'input_errored_class':FieldArg("cssclass", ''),
                        'set_input_accepted':FieldArg("boolean", False, jsonset=True),
                        'set_input_errored':FieldArg("boolean", False, jsonset=True),
                        'labeldiv_class':FieldArg("cssclass", ''),
                        'labeldiv_style':FieldArg("cssstyle", ''),
                        'label':FieldArg("text", 'Your input:'),
                        'label_class':FieldArg("cssclass", ''),
                        'label_style':FieldArg("cssstyle", ''),
                        'inputdiv_class':FieldArg("cssclass", ''),
                        'inputdiv_style':FieldArg("cssstyle", ''),
                        'input_text':FieldArg("text", '', valdt=True, jsonset=True),
                        'input_class':FieldArg("cssclass", ''),
                        'input_style':FieldArg("cssstyle", ''),
                        'input_disabled_class':FieldArg("cssclass", ''),
                        'size':FieldArg("text", ''),
                        'maxlength':FieldArg("text", ''),
                        'redstar':FieldArg("boolean", False),
                        'redstar_class':FieldArg("cssclass", ''),
                        'redstar_style':FieldArg("cssstyle", ''),
                        'disabled':FieldArg("boolean", False, jsonset=True),
                        'required':FieldArg("boolean", False),
                        'type':FieldArg("text", 'text'),
                        'pattern':FieldArg("text", ''),
                        'title':FieldArg("text", '')
                       }
    def __init__(self, name=None, brief='', **field_args):
        """
        input_accepted_class: A class which can be set on the input field
        input_errored_class: A class which can be set on the input field
        set_input_accepted: If True, input_accepted_class will be set on the input field
        set_errored_accepted: If True, input_errored_class will be set on the input field
        labeldiv_class: The css class of the div containing the label
        label: The text appearing in a label tag to the left of the input field
        label_class: The css class of the label to the left of the input field
        inputdiv_class: The css class of the div containing the input field
        input_text: The default text in the text input field, field name used as name attribute
        input_class: The css class of the input field
        input_disabled_class: css class of the input field which replaces input_class if disabled is True
        size: The number of characters appearing in the text input area
        maxlength: The maximum number of characters accepted in the text area
        redstar: If True a red asterix is shown by the side of the input field
        disabled: Set True if the field is to be disabled
        required: Set True to put the 'required' flag into the input field
        type: the type set on the input field, such as text or email
        pattern: regular expression pattern
        title: helps describe the pattern
        """
        Widget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "div"
        # label div at 0
        self[0] = tag.Part(tag_name="div")
        # the label
        self[0][0] = tag.Part(tag_name="label", hide_if_empty=True)
        # input div at 1
        self[1] = tag.Part(tag_name="div")
        # the text input field
        self[1][0] = tag.ClosedPart(tag_name="input", attribs ={"type":"text"})


    def _build(self, page, ident_list, environ, call_data, lang):
        "build the input field"
        # label
        self[0].set_class_style(self.wf.labeldiv_class, self.wf.labeldiv_style)
        self[0][0].set_class_style(self.wf.label_class, self.wf.label_style)

        if self.wf.label:
            self[0][0][0] = self.wf.label
        # input div
        self[1].set_class_style(self.wf.inputdiv_class, self.wf.inputdiv_style)

        # set an id in the input field for the 'label for' tag
        # any label:value added to self.jlabels will be set in a javascript fieldvalues attribute for the widget
        self.jlabels['input_id'] = self[1][0].insert_id()
        self[0][0].attribs['for'] = self.jlabels['input_id']

        if self.wf.input_accepted_class:
            self.jlabels['input_accepted_class'] = self.wf.input_accepted_class
        if self.wf.input_errored_class:
            self.jlabels['input_errored_class'] = self.wf.input_errored_class
        if self.wf.input_class:
            self.jlabels['input_class'] = self.wf.input_class
        if self.wf.input_disabled_class:
            self.jlabels['input_disabled_class'] = self.wf.input_disabled_class

        # input field
        self[1][0].attribs.update({"name":self.get_formname('input_text'), "value":self.wf.input_text})

        if self.wf.size:
            self[1][0].attribs["size"] = self.wf.size
        if self.wf.maxlength:
            self[1][0].attribs["maxlength"] = self.wf.maxlength
        if self.wf.required:
            self[1][0].attribs["required"] = "required"
        if self.wf.type:
            self[1][0].attribs["type"] = self.wf.type
        if self.wf.pattern:
            self[1][0].attribs["pattern"] = self.wf.pattern
        if self.wf.title:
            self[1][0].attribs["title"] = self.wf.title

        if self.wf.input_class:
            input_class = self.wf.input_class
        else:
            input_class = ''
        if self.wf.disabled:
            self[1][0].attribs["disabled"] = "disabled"
            if self.wf.input_disabled_class:
                input_class = self.wf.input_disabled_class
        if self.wf.set_input_errored and self.wf.input_errored_class:
            if input_class:
                input_class = input_class + ' ' + self.wf.input_errored_class
            else:
                input_class = self.wf.input_errored_class
        elif self.wf.set_input_accepted and self.wf.input_accepted_class:
            if input_class:
                input_class = input_class + ' ' + self.wf.input_accepted_class
            else:
                input_class = self.wf.input_accepted_class

        self[1][0].set_class_style(input_class, self.wf.input_style)

        # redstar
        if self.wf.redstar:
            self[1][1] = tag.Part(tag_name="span")
            self[1][1].set_class_style(self.wf.redstar_class, self.wf.redstar_style)
            self[1][1][0] = '*'


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
  <div> <!-- with class set to labeldiv_class -->
    <label> <!-- with class set to label_class -->
            <!-- content set to label -->
    </label>
  </div>
  <div> <!-- with class set to inputdiv_class -->
    <input type='text' /> <!-- with attributes set appropriately -->
    <span>*</span> <!-- shown if redstar is True -->
  </div>
</div>"""



class SubmitTextInput1(Widget):
    """Defines a form with a text input field, and four hidden fields"""

    # js_validators is a class attribute, True if javascript validation is enabled
    js_validators=True

    error_location = (0,0,0)

    arg_descriptions = {'label':FieldArg("text", 'Your input:'),
                        'label_class':FieldArg("cssclass", ''),
                        'label_style':FieldArg("cssstyle", ''),
                        'action_json':FieldArg("url", ''),
                        'action':FieldArg("url", ''),
                        'hidden_field1':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field2':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field3':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field4':FieldArg("text", '', valdt=True, jsonset=True),
                        'target':FieldArg("text",''),
                        'button_text':FieldArg("text",'Submit'),
                        'button_class':FieldArg("cssclass", ''),
                        'inputdiv_class':FieldArg("cssclass", ''),
                        'inputandbutton_class':FieldArg("cssclass", ''),
                        'inputandbutton_style':FieldArg("cssstyle", ''),
                        'error_class':FieldArg("cssclass", ''),
                        'input_text':FieldArg("text", '', valdt=True, jsonset=True),
                        'size':FieldArg("text", ''),
                        'maxlength':FieldArg("text", ''),
                        'required':FieldArg("boolean", False),
                        'type':FieldArg("text", 'text'),
                        'pattern':FieldArg("text", ''),
                        'title':FieldArg("text", ''),
                        'input_accepted_class':FieldArg("cssclass", ''),
                        'input_errored_class':FieldArg("cssclass", ''),
                        'input_class':FieldArg("cssclass", ''),
                        'set_input_accepted':FieldArg("boolean", False, jsonset=True),
                        'set_input_errored':FieldArg("boolean", False, jsonset=True),
                        'hide':FieldArg("boolean", False, jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        label: The text displayed to the left of the text input field
        label_class: The css class of the label
        action_json:  if a value set, and client has jscript enabled, this is the page ident, label, url this button links to, expects a json page back
        action: The page ident, label, url this button links to, overridden if action_json is set.
        hidden_field1: A hidden field value, leave blank if unused, name used as the get field name
        hidden_field2: A second hidden field value, leave blank if unused, name used as the get field name
        hidden_field3: A third hidden field value, leave blank if unused, name used as the get field name
        hidden_field4: A fourth hidden field value, leave blank if unused, name used as the get field name
        target: if given, the target attribute will be set
        button_text: The text on the button
        button_class: The class given to the button tag
        inputdiv_class: the class attribute of the div which contains the label, input text and button
        inputandbutton_class: the class attribute of the span which contains the input text and button
        error_class: The class applied to the paragraph containing the error message on error.
        input_text: The default text in the text input field, field name used as the name attribute
        size: The number of characters appearing in the text input area
        maxlength: The maximum number of characters accepted in the text area
        required: Set True to put the 'required' flag into the input field
        type: the type set on the input field, such as text or email
        pattern: regular expression pattern
        title: helps describe the pattern
        input_accepted_class: A class which can be set on the input field
        input_errored_class: A class which can be set on the input field
        input_class: Class set on the input field
        set_input_accepted: If True, input_accepted_class will be set on the input field
        set_errored_accepted: If True, input_errored_class will be set on the input field
        hide: If True, widget is hidden
        """
        Widget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "div"
        # error div at 0
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''
        # The form
        self[1] = tag.Part(tag_name='form', attribs={"role":"form", "method":"post"})

        # div containing label, input text and button
        self[1][0] = tag.Part(tag_name='div')
        # the label
        self[1][0][0] = tag.Part(tag_name="label", hide_if_empty=True)
        # span containing input text and button
        self[1][0][1] = tag.Part(tag_name='span')
        # the text input field
        self[1][0][1][0] = tag.ClosedPart(tag_name="input", attribs ={"type":"text"})
        # the submit button
        self[1][0][1][1] = tag.Part(tag_name="button", attribs ={"type":"submit"})
        self[1][0][1][1][0] = "Submit"


    def _build(self, page, ident_list, environ, call_data, lang):
        "build the form"
        if self.wf.target:
            self[1].attribs["target"] = self.wf.target
        # Hides widget if no error and hide is True
        self.widget_hide(self.wf.hide)

        # any label:value added to self.jlabels will be set in a javascript fieldvalues attribute for the widget
        jsonurl = self.get_url(self.wf.action_json)
        if jsonurl:
            self.jlabels['url'] = jsonurl

        if self.wf.error_class:
            self[0].attribs["class"] = self.wf.error_class
        if self.error_status:
            del self[0].attribs["style"]
        if not self.wf.action:
            # setting self._error replaces the entire tag
            self._error = "Warning: No form action"
            return
        actionurl = self.get_url(self.wf.action)
        if not actionurl:
            # setting self._error replaces the entire tag
            self._error = "Warning: broken link"
            return
        # update the action of the form
        self[1].attribs["action"] = actionurl
        # the div holding label, input text and button
        if self.wf.inputdiv_class:
            self[1][0].attribs["class"] = self.wf.inputdiv_class

        self[1][0][0].set_class_style(self.wf.label_class, self.wf.label_style)
        if self.wf.label:
            self[1][0][0][0] = self.wf.label

        # the span holding input text and button
        self[1][0][1].set_class_style(self.wf.inputandbutton_class, self.wf.inputandbutton_style)

        # set an id in the input field for the 'label for' tag
        self.jlabels['input_id'] = self[1][0][1][0].insert_id()

        self[1][0][1][0].attribs.update({"name":self.get_formname('input_text'), "value":self.wf.input_text})

        if self.wf.size:
            self[1][0][1][0].attribs["size"] = self.wf.size
        if self.wf.maxlength:
            self[1][0][1][0].attribs["maxlength"] = self.wf.maxlength
        if self.wf.required:
            self[1][0][1][0].attribs["required"] = "required"
        if self.wf.type:
            self[1][0][1][0].attribs["type"] = self.wf.type
        if self.wf.pattern:
            self[1][0][1][0].attribs["pattern"] = self.wf.pattern
        if self.wf.title:
            self[1][0][1][0].attribs["title"] = self.wf.title

        if self.wf.input_accepted_class:
            self.jlabels['input_accepted_class'] = self.wf.input_accepted_class
        if self.wf.input_errored_class:
            self.jlabels['input_errored_class'] = self.wf.input_errored_class

        if self.wf.input_class:
            input_class = self.wf.input_class
        else:
            input_class = ''

        if self.error_status and self.wf.input_errored_class:
            if input_class:
                input_class = input_class + ' ' + self.wf.input_errored_class
            else:
                input_class = self.wf.input_errored_class
        elif self.wf.set_input_errored and self.wf.input_errored_class:
            if input_class:
                input_class = input_class + ' ' + self.wf.input_errored_class
            else:
                input_class = self.wf.input_errored_class
        elif self.wf.set_input_accepted and self.wf.input_accepted_class:
            if input_class:
                input_class = input_class + ' ' + self.wf.input_accepted_class
            else:
                input_class = self.wf.input_accepted_class

        if input_class:
            self[1][0][1][0].attribs["class"] = input_class

        # set the label 'for' attribute
        self[1][0][0].attribs['for'] = self.jlabels['input_id']

        # submit button
        if self.wf.button_class:
            self[1][0][1][1].attribs["class"] = self.wf.button_class
        if self.wf.button_text:
            self[1][0][1][1][0] = self.wf.button_text

        # add ident and four hidden fields to the form at self[1]
        self.add_hiddens(self[1], page)


    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets a submit event handler"""
        ident=self.get_id()
        return f"""  $("#{ident} form").on("submit input", function(e) {{
    SKIPOLE.widgets["{ident}"].eventfunc(e);
    }});
"""

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div> <!-- with widget id and class widget_class -->
  <div>  <!-- div hidden when no error is displayed, with class set to error_class on error -->
    <p> <!-- error message appears in this paragraph --> </p>
  </div>
  <form role="form" method="post"> <!-- action attribute set to action field -->
    <div> <!-- class attribute set to inputdiv_class -->
      <label> <!-- with class set to label_class and content to label, for set to input text id -->
      </label>
      <span>  <!-- class attribute set to inputandbutton_class -->
          <input type="text" />  <!-- class set to input_class -->
            <!-- input text value set to input_text -->
          <button type="submit"> <!-- with class set to button_class -->
            <!-- button_text -->
          </button>
      </span>
    </div>
    <!-- hidden input fields -->
  </form>
</div>"""



class SubmitTextInput3(Widget):
    """Defines paragraphs followed by a form with a text input field"""

    # js_validators is a class attribute, True if javascript validation is enabled
    js_validators=True

    error_location = (0,0,0)

    arg_descriptions = {
                        'show_para1':FieldArg("boolean", True, jsonset=True),
                        'para_text':FieldArg("text", '', jsonset=True),
                        'show_para2':FieldArg("boolean", True, jsonset=True),
                        'textblock_ref':FieldArg("textblock_ref", ""),
                        'label':FieldArg("text", 'Your input:'),
                        'label_class':FieldArg("cssclass", ''),
                        'label_style':FieldArg("cssstyle", ''),
                        'button_label':FieldArg("text", 'Please submit:'),
                        'button_label_class':FieldArg("cssclass", ''),
                        'button_label_style':FieldArg("cssstyle", ''),
                        'action':FieldArg("url", ''),
                        'action_json':FieldArg("url", ''),
                        'hidden_field1':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field2':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field3':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field4':FieldArg("text", '', valdt=True, jsonset=True),
                        'target':FieldArg("text",''),
                        'button_text':FieldArg("text",'Submit'),
                        'button_class':FieldArg("cssclass", ''),
                        'buttondiv_class':FieldArg("cssclass", ''),
                        'outer_class':FieldArg("cssclass", ''),
                        'inputdiv_class':FieldArg("cssclass", ''),
                        'error_class':FieldArg("cssclass", ""),
                        'input_text':FieldArg("text", '', valdt=True, jsonset=True),
                        'size':FieldArg("text", ''),
                        'maxlength':FieldArg("text", ''),
                        'required':FieldArg("boolean", False),
                        'type':FieldArg("text", 'text'),
                        'pattern':FieldArg("text", ''),
                        'title':FieldArg("text", ''),
                        'input_accepted_class':FieldArg("cssclass", ''),
                        'input_errored_class':FieldArg("cssclass", ''),
                        'input_class':FieldArg("cssclass", ''),
                        'set_input_accepted':FieldArg("boolean", False, jsonset=True),
                        'set_input_errored':FieldArg("boolean", False, jsonset=True),
                        'hide':FieldArg("boolean", False, jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        show_para1: True if the paragraph is to be visible, False if not
        para_text: The text appearing in the paragraph
        show_para2: True if the paragraph is to be visible, False if not
        textblock_ref: The reference of the TextBlock appearing in the second paragraph
        label: The text displayed to the left of the text input field
        label_class: The css class of the label
        button_label: The text displayed to the left of the button
        button_label_class: The css class of the label to the left of the button
        action_json:  if a value set, and client has jscript enabled, this is the page ident, label, url this button links to, expects a json page back
        action: The page ident, label, url this button links to
        hidden_field1: A hidden field value, leave blank if unused, name used as the get field name
        hidden_field2: A second hidden field value, leave blank if unused, name used as the get field name
        hidden_field3: A third hidden field value, leave blank if unused, name used as the get field name
        hidden_field4: A fourth hidden field value, leave blank if unused, name used as the get field name
        target: if given, the target attribute will be set
        button_text: The text on the button
        buttondiv_class: the class attribute of the div which contains the button
        button_class: the class attribute of the button
        outer_class: the class attribute of the outside div
        inputdiv_class: the class attribute of the tag which contains the label and input text
        error_class: The class applied to the paragraph containing the error message on error.
        input_text: The default text in the text input field, field name used as the name attribute
        size: The number of characters appearing in the text input area
        maxlength: The maximum number of characters accepted in the text area
        required: Set True to put the 'required' flag into the input field
        type: the type set on the input field, such as text or email
        pattern: regular expression pattern
        title: helps describe the pattern
        input_accepted_class: A class which can be set on the input field
        input_errored_class: A class which can be set on the input field
        input_class: Class set on the input field
        set_input_accepted: If True, input_accepted_class will be set on the input field
        set_errored_accepted: If True, input_errored_class will be set on the input field
        hide: If True, widget is hidden
        """
        Widget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "div"
        # error div at 0
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''
        # div containing paragraphs
        self[1] = tag.Part(tag_name='div')
        # paragraph one and two
        self[1][0] = tag.Part(tag_name='p')
        self[1][0][0] =''
        self[1][1] = tag.Part(tag_name='p')
        self[1][1][0] = ''

        # The form
        self[2] = tag.Part(tag_name='form', attribs={"role":"form", "method":"post"})

        # div containing label and input text
        self[2][0] = tag.Part(tag_name='div')
        # the label
        self[2][0][0] = tag.Part(tag_name="label", hide_if_empty=True)
        # the text input field
        self[2][0][1] = tag.ClosedPart(tag_name="input", attribs ={"type":"text"})

        # div containing button
        self[2][1] = tag.Part(tag_name='div')
        # the label on the left of the button
        self[2][1][0] = tag.Part(tag_name="label", hide_if_empty=True)
        # the submit button
        self[2][1][1] = tag.ClosedPart(tag_name="input", attribs={"type":"submit"})


    def _build(self, page, ident_list, environ, call_data, lang):
        "build the form"

        if self.wf.target:
            self[2].attribs["target"] = self.wf.target
        # Hides widget if no error and hide is True
        self.widget_hide(self.wf.hide)

        # any label:value added to self.jlabels will be set in a javascript fieldvalues attribute for the widget
        jsonurl = self.get_url(self.wf.action_json)
        if jsonurl:
            self.jlabels['url'] = jsonurl

        if self.wf.error_class:
            self[0].attribs["class"] = self.wf.error_class
        if self.error_status:
            del self[0].attribs["style"]
        if not self.wf.action:
            # setting self._error replaces the entire tag
            self._error = "Warning: No form action"
            return
        actionurl = self.get_url(self.wf.action)
        if not actionurl:
            # setting self._error replaces the entire tag
            self._error = "Warning: broken link"
            return
        # update the action of the form
        self[2].attribs["action"] = actionurl

        # the class of the outer div
        self[1].set_class_style(self.wf.outer_class)


        # set paragraphs
        if not self.wf.show_para1:
            self[1][0].attribs["style"] = "display:none;"
        self[1][0][0] = self.wf.para_text
        if not self.wf.show_para2:
            self[1][1].attribs["style"] = "display:none;"
        # define the textblock
        tblock = self.wf.textblock_ref
        if tblock:
            tblock.proj_ident = page.proj_ident
            self[1][1][0] = tblock

        # the div holding label and input text
        self[2][0].set_class_style(self.wf.inputdiv_class)

        if self.wf.label:
            self[2][0][0][0] = self.wf.label

        self[2][0][0].set_class_style(self.wf.label_class, self.wf.label_style)

        # set an id in the input field for the 'label for' tag
        self.jlabels['input_id'] = self[2][0][1].insert_id()

        self[2][0][1].attribs.update({"name":self.get_formname('input_text'), "value":self.wf.input_text})

        if self.wf.size:
            self[2][0][1].attribs["size"] = self.wf.size
        if self.wf.maxlength:
            self[2][0][1].attribs["maxlength"] = self.wf.maxlength
        if self.wf.required:
            self[2][0][1].attribs["required"] = "required"
        if self.wf.type:
            self[2][0][1].attribs["type"] = self.wf.type
        if self.wf.pattern:
            self[2][0][1].attribs["pattern"] = self.wf.pattern
        if self.wf.title:
            self[2][0][1].attribs["title"] = self.wf.title

        if self.wf.input_accepted_class:
            self.jlabels['input_accepted_class'] = self.wf.input_accepted_class
        if self.wf.input_errored_class:
            self.jlabels['input_errored_class'] = self.wf.input_errored_class

        if self.wf.input_class:
            input_class = self.wf.input_class
        else:
            input_class = ''

        if self.error_status and self.wf.input_errored_class:
            if input_class:
                input_class = input_class + ' ' + self.wf.input_errored_class
            else:
                input_class = self.wf.input_errored_class
        elif self.wf.set_input_errored and self.wf.input_errored_class:
            if input_class:
                input_class = input_class + ' ' + self.wf.input_errored_class
            else:
                input_class = self.wf.input_errored_class
        elif self.wf.set_input_accepted and self.wf.input_accepted_class:
            if input_class:
                input_class = input_class + ' ' + self.wf.input_accepted_class
            else:
                input_class = self.wf.input_accepted_class

        if input_class:
            self[2][0][1].attribs["class"] = input_class

        # set the label 'for' attribute
        self[2][0][0].attribs['for'] = self.jlabels['input_id']

        # the div holding button
        if self.wf.buttondiv_class:
            self[2][1].attribs["class"] = self.wf.buttondiv_class

        if self.wf.button_label:
            self[2][1][0][0] = self.wf.button_label

        self[2][1][0].set_class_style(self.wf.button_label_class, self.wf.button_label_style)

        # submit button
        self[2][1][1].attribs["value"] = self.wf.button_text
        # set an id in the submit button for the 'label for' tag
        submit_id = self[2][1][1].insert_id()
        # the button class
        if self.wf.button_class:
            self[2][1][1].attribs["class"] = self.wf.button_class

        # set the label 'for' attribute
        self[2][1][0].attribs['for'] = submit_id

        # add ident and four hidden fields
        self.add_hiddens(self[2], page)


    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets a submit event handler"""
        ident=self.get_id()
        return f"""  $("#{ident} form").on("submit input", function(e) {{
    SKIPOLE.widgets["{ident}"].eventfunc(e);
    }});
"""

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div> <!-- with widget id and class widget_class -->
  <div> <!-- normally hidden div, with class error_class -->
    <p> <!-- Any error text appears here --> </p>
  </div>
  <div> <!-- class attribute set to outer_class -->
    <p><!-- Paragraph 1, hidden if show_para1 is False -->
      <!-- Containing para_text -->
    </p>
    <p><!-- Paragraph 2, hidden if show_para2 is False -->
      <!-- Contains TextBlock -->
    </p>
  <div>
  <form method="post"> <!-- action attribute set to action field -->
    <div> <!-- class attribute set to inputdiv_class -->
      <label> <!-- with class set to label_class and content to label -->
      </label>
      <input type=\"text\" /> <!-- input text value set to input_text -->
    </div>
    <div> <!-- class attribute set to buttondiv_class -->
      <label> <!-- with class set to button_label_class and content to button_label -->
      </label>
      <input type=\"submit\" /> <!-- button value set to button_text and class to button_class -->
    </div>
    <!-- hidden input fields -->
  </form>
</div>"""


class TwoInputsSubmit1(Widget):
    """Defines a form containing two text input fields."""

    error_location = (0,0,0)

    # js_validators is a class attribute, True if javascript validation is enabled
    js_validators=True

    arg_descriptions = {'action':FieldArg("url", ''),
                        'action_json':FieldArg("url", ''),
                        'error_class':FieldArg("cssclass", ""),
                        'inputdiv_class':FieldArg("cssclass", ''),
                        'label':FieldArg("text", 'Your input:'),
                        'label_class':FieldArg("cssclass", ''),
                        'label_style':FieldArg("cssstyle", ''),
                        'hidden_field1':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field2':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field3':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field4':FieldArg("text", '', valdt=True, jsonset=True),
                        'button_text':FieldArg("text", 'Submit'),
                        'button_class':FieldArg("cssclass", ''),
                        'input_text1':FieldArg("text", '', valdt=True, jsonset=True),
                        'size1':FieldArg("text", ''),
                        'maxlength1':FieldArg("text", ''),
                        'disabled1':FieldArg("boolean", False),
                        'required1':FieldArg("boolean", False),
                        'type1':FieldArg("text", 'text'),
                        'pattern1':FieldArg("text", ''),
                        'title1':FieldArg("text", ''),
                        'set_input_accepted1':FieldArg("boolean", False, jsonset=True),
                        'set_input_errored1':FieldArg("boolean", False, jsonset=True),
                        'input_text2':FieldArg("text", '', valdt=True, jsonset=True),
                        'size2':FieldArg("text", ''),
                        'maxlength2':FieldArg("text", ''),
                        'disabled2':FieldArg("boolean", False),
                        'required2':FieldArg("boolean", False),
                        'type2':FieldArg("text", 'text'),
                        'pattern2':FieldArg("text", ''),
                        'title2':FieldArg("text", ''),
                        'set_input_accepted2':FieldArg("boolean", False, jsonset=True),
                        'set_input_errored2':FieldArg("boolean", False, jsonset=True),
                        'input_class1':FieldArg("cssclass", ''),
                        'input_class2':FieldArg("cssclass", ''),
                        'input_accepted_class':FieldArg("cssclass", ''),
                        'input_errored_class':FieldArg("cssclass", '')
                       }


    def __init__(self, name=None, brief='', **field_args):
        """
        action: The page ident this button links to
        action_json:  if a value set, and client has jscript enabled, this is the page ident, label, url this button links to, expects a json page back
        error_class: The class applied to the paragraph containing the error message on error.
        inputdiv_class: the class attribute of the tag which contains the label and inputs
        label: The text displayed to the left of the text input field
        label_class: The css class of the label
        hidden_field1: A hidden field value, leave blank if unused, name used as the get field name
        hidden_field2: A second hidden field value, leave blank if unused, name used as the get field name
        hidden_field3: A third hidden field value, leave blank if unused, name used as the get field name
        hidden_field4: A fourth hidden field value, leave blank if unused, name used as the get field name
        button_text: The text on the button
        button_class: the class attribute of the button
        input_accepted_class: A class which can be set on each input field
        input_errored_class: A class which can be set on each input field
        input_text1: The default text in the first text input field, field name used as the name attribute
        input_class1: Class set on the first input field
        size1: The number of characters appearing in the first text input area
        maxlength1: The maximum number of characters accepted in the first text area
        disabled1: Set True if the first input field is to be disabled
        required1: Set True to put the 'required' flag into the first input field
        type1: the type set on the first input field, such as text or email
        pattern1: regular expression pattern for first input field
        title1: helps describe the pattern of the first input field
        input_text2: The default text in the second text input field, field name used as the name attribute
        input_class2: Class set on the second input field
        size2: The number of characters appearing in the second text input area
        maxlength2: The maximum number of characters accepted in the second text area
        disabled2: Set True if the second field is to be disabled
        required2: Set True to put the 'required' flag into the second input field
        type2: the type set on the second input field, such as text or email
        pattern2: regular expression pattern for the second input field
        title2: helps describe the pattern of the second input field
        """
        Widget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "div"
        # error div at 0
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''
        # The form
        self[1] = tag.Part(tag_name='form', attribs={"role":"form", "method":"post"})
        # div containing label, input text and button
        self[1][0] = tag.Part(tag_name='div')

        # the label
        self[1][0][0] = tag.Part(tag_name="label", hide_if_empty=True)
        # the first text input field
        self[1][0][1] = tag.ClosedPart(tag_name="input", attribs ={"type":"text"})
        # the second text input field
        self[1][0][2] = tag.ClosedPart(tag_name="input", attribs ={"type":"text"})
        # the submit button
        self[1][0][3] = tag.ClosedPart(tag_name="input", attribs={"type":"submit"})


    def _build(self, page, ident_list, environ, call_data, lang):
        "build the form"

        if not self.wf.action:
            # setting self._error replaces the entire tag
            self._error = "Warning: No form action"
            return
        actionurl = self.get_url(self.wf.action)
        if not actionurl:
            # setting self._error replaces the entire tag
            self._error = "Warning: broken link"
            return
        # update the action of the form
        self[1].attribs["action"] = actionurl

        # any label:value added to self.jlabels will be set in a javascript fieldvalues attribute for the widget
        jsonurl = self.get_url(self.wf.action_json)
        if jsonurl:
            self.jlabels['url'] = jsonurl

        if self.wf.error_class:
            self[0].attribs["class"] = self.wf.error_class
        if self.error_status:
            del self[0].attribs["style"]


        # the div holding label, input text and button
        if self.wf.inputdiv_class:
            self[1][0].attribs["class"] = self.wf.inputdiv_class

        # set up label
        if self.wf.label:
            self[1][0][0][0] = self.wf.label
        self[1][0][0].set_class_style(self.wf.label_class, self.wf.label_style)

        # first input field
        self[1][0][1].attribs.update({"name":self.get_formname('input_text1'), "value":self.wf.input_text1})

        if self.wf.size1:
            self[1][0][1].attribs["size"] = self.wf.size1
        if self.wf.maxlength1:
            self[1][0][1].attribs["maxlength"] = self.wf.maxlength1
        if self.wf.disabled1:
            self[1][0][1].attribs["disabled"] = "disabled"
        if self.wf.required1:
            self[1][0][1].attribs["required"] = "required"
        if self.wf.type1:
            self[1][0][1].attribs["type"] = self.wf.type1
        if self.wf.pattern1:
            self[1][0][1].attribs["pattern"] = self.wf.pattern1
        if self.wf.title1:
            self[1][0][1].attribs["title"] = self.wf.title1

        if self.wf.input_class1:
            input_class1 = self.wf.input_class1
        else:
            input_class1 = ''

        if self.error_status and self.wf.input_errored_class:
            if input_class1:
                input_class1 = input_class1 + ' ' + self.wf.input_errored_class
            else:
                input_class1 = self.wf.input_errored_class
        elif self.wf.set_input_errored1 and self.wf.input_errored_class:
            if input_class1:
                input_class1 = input_class1 + ' ' + self.wf.input_errored_class
            else:
                input_class1 = self.wf.input_errored_class
        elif self.wf.set_input_accepted1 and self.wf.input_accepted_class:
            if input_class1:
                input_class1 = input_class1 + ' ' + self.wf.input_accepted_class
            else:
                input_class1 = self.wf.input_accepted_class

        if input_class1:
            self[1][0][1].attribs["class"] = input_class1


        # second input field
        self[1][0][2].attribs.update({"name":self.get_formname('input_text2'), "value":self.wf.input_text2})

        if self.wf.size2:
            self[1][0][2].attribs["size"] = self.wf.size2
        if self.wf.maxlength2:
            self[1][0][2].attribs["maxlength"] = self.wf.maxlength2
        if self.wf.disabled2:
            self[1][0][2].attribs["disabled"] = "disabled"
        if self.wf.required2:
            self[1][0][2].attribs["required"] = "required"
        if self.wf.type2:
            self[1][0][2].attribs["type"] = self.wf.type2
        if self.wf.pattern2:
            self[1][0][2].attribs["pattern"] = self.wf.pattern2
        if self.wf.title2:
            self[1][0][2].attribs["title"] = self.wf.title2

        if self.wf.input_class2:
            input_class2 = self.wf.input_class2
        else:
            input_class2 = ''

        if self.error_status and self.wf.input_errored_class:
            if input_class2:
                input_class2 = input_class2 + ' ' + self.wf.input_errored_class
            else:
                input_class2 = self.wf.input_errored_class
        elif self.wf.set_input_errored2 and self.wf.input_errored_class:
            if input_class2:
                input_class2 = input_class2 + ' ' + self.wf.input_errored_class
            else:
                input_class2 = self.wf.input_errored_class
        elif self.wf.set_input_accepted2 and self.wf.input_accepted_class:
            if input_class2:
                input_class2 = input_class2 + ' ' + self.wf.input_accepted_class
            else:
                input_class2 = self.wf.input_accepted_class

        if input_class2:
            self[1][0][2].attribs["class"] = input_class2

        if self.wf.input_accepted_class:
            self.jlabels['input_accepted_class'] = self.wf.input_accepted_class
        if self.wf.input_errored_class:
            self.jlabels['input_errored_class'] = self.wf.input_errored_class


        # submit button
        self[1][0][3].attribs["value"] = self.wf.button_text
        # the button class
        if self.wf.button_class:
            self[1][0][3].attribs["class"] = self.wf.button_class

        # set an id in the first input field for the 'label for' tag
        submit_id = self[1][0][1].insert_id()
        # set the label 'for' attribute
        self[1][0][0].attribs['for'] = submit_id

        # add ident and four hidden fields
        self.add_hiddens(self[1], page)


    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets a submit event handler"""
        ident=self.get_id()
        return f"""  $("#{ident} form").on("submit input", function(e) {{
    SKIPOLE.widgets["{ident}"].eventfunc(e);
    }});
"""


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div> <!-- with widget id and class widget_class -->
  <div> <!-- normally hidden div, with class error_class -->
    <p> <!-- Any error text appears here --> </p>
  </div>
  <form method="post"> <!-- action attribute set to action field -->
    <div> <!-- class attribute set to inputdiv_class -->
      <label> <!-- with class set to label_class and content to label -->
      </label>
      <input type=\"text\" /> <!-- input text value set to input_text1, class to input_class1 -->
      <input type=\"text\" /> <!-- input text value set to input_text2, class to input_class2 -->
      <input type=\"submit\" /> <!-- button value set to button_text and class to button_class -->
    </div>
    <!-- hidden input fields -->
  </form>
</div>"""


class SubmitDict1(Widget):
    """Defines a form with a list of input fields followed by a single submit button
       to allow a client to input a dictionary"""

    display_errors = False

    # js_validators is a class attribute, True if javascript validation is enabled
    js_validators=True

    arg_descriptions = {'action':FieldArg("url", ''),
                        'button_text':FieldArg("text", "Submit"),
                        'button_class':FieldArg("cssclass", ''),
                        'input_dict':FieldArgDict('text', valdt=True, senddict=True),
                        'input_class':FieldArg("cssclass", ''),
                        'inputdiv_class':FieldArg("cssclass", ''),
                        'inputdiv_style':FieldArg("cssstyle", ''),
                        'ul_class':FieldArg("cssclass", ""),
                        'ul_style':FieldArg("cssstyle", ""),
                        'li_class':FieldArg("cssclass", ""),
                        'li_style':FieldArg("cssstyle", ""),
                        'hidden_field1':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field2':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field3':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field4':FieldArg("text", '', valdt=True, jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        action: The label or ident of the action page
        button_text: text appearing on the button
        button_class: The css class of the button
        input_dict: an ordered dictionary, keys used in the field name, values are the text input
        input_class: Class set on each input field
        ul_class: Unordered list class
        ul_style: Style applied to ul element
        li_class: list element class
        li_style: Style applied to li element
        hidden_field1: A hidden field value, leave blank if unused, name used as the get field name
        hidden_field2: A second hidden field value, leave blank if unused, name used as the get field name
        hidden_field3: A third hidden field value, leave blank if unused, name used as the get field name
        hidden_field4: A fourth hidden field value, leave blank if unused, name used as the get field name
        """
        Widget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "div"
        # The form
        self[0] = tag.Part(tag_name='form', attribs={"method":"post"})
        self[0][0] = tag.Part(tag_name='ul')
        # the submit button in a div
        self[0][1] = tag.Part(tag_name="div")
        self[0][1][0] = tag.ClosedPart(tag_name="input")

    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the form and list"

        input_dict = self.wf.input_dict
        if not input_dict:
            self.show = False
            return

        if not self.wf.action:
            # setting self._error replaces the entire tag
            self._error = "Warning: No form action"
            return
        actionurl = self.get_url(self.wf.action)
        if not actionurl:
            # setting self._error replaces the entire tag
            self._error = "Warning: broken link"
            return
        # update the action of the form
        self[0].attribs["action"] = actionurl

        self[0][0].set_class_style(self.wf.ul_class, self.wf.ul_style)

        li_class =  self.wf.li_class

        input_name = self.get_formname("input_dict") + '-'

        if self.wf.input_class:
            input_class = self.wf.input_class
        else:
            input_class = ''

        linumber = -1
        for key, value in input_dict.items():
            keyed_name = input_name + key
            if input_class:
                li_input = tag.ClosedPart(tag_name="input",
                                          attribs ={"name":keyed_name,
                                                    "value":value,
                                                    "type":"text",
                                                    "class":input_class})
            else:
                li_input = tag.ClosedPart(tag_name="input",
                                          attribs ={"name":keyed_name,
                                                    "value":value,
                                                    "type":"text"})
            linumber += 1
            if li_class:
                self[0][0][linumber] = tag.Part(tag_name="li", attribs ={"class":li_class})
            else:
                self[0][0][linumber] = tag.Part(tag_name="li")
            if self.wf.li_style:
                self[0][0][linumber].attribs["style"] = self.wf.li_style

            self[0][0][linumber][0] = li_input

        # list done, now for submit button

        # the div holding button
        self[0][1].set_class_style(self.wf.inputdiv_class, self.wf.inputdiv_style)

        if self.wf.button_class:
            self[0][1][0].attribs = {"value":self.wf.button_text, "type":"submit", "class":self.wf.button_class}
        else:
            self[0][1][0].attribs = {"value":self.wf.button_text, "type":"submit"}

        # add ident and four hidden fields
        self.add_hiddens(self[0], page)

    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets a submit event handler for javascript validation"""
        ident=self.get_id()
        return f"""  $("#{ident} form").on("submit input", function(e) {{
    SKIPOLE.widgets["{ident}"].eventfunc(e);
    }});
"""


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div> <!-- with widget id and class widget_class -->
  <form method=\"post\"> <!-- action attribute set to action field -->
    <ul> <!-- with class set to ul_class and style to ul_style -->
      <!-- a li is created for each value in the input_dict dictionary -->
      <li>  <!-- with class set to li_class and style to li_style -->
        <input type="text" /> <!-- with CSS class input_class -->
             <!-- and with input name 'widgetname:input_dict-keyname' -->
      </li>
    </ul>
    <div> <!-- class attribute set to inputdiv_class, style to inputdiv_style -->
      <input type="submit" /> <!-- with value button_text, and CSS class button_class -->
    </div>
    <!-- hidden fields -->
  </form>
</div>"""



class SubmitTextInput2(Widget):
    """Defines a form with a text input field, and four hidden fields
       Can send session or local storage values."""

    # js_validators is a class attribute, True if javascript validation is enabled
    js_validators=True

    error_location = (0,0,0)

    arg_descriptions = {'label':FieldArg("text", 'Your input:'),
                        'label_class':FieldArg("cssclass", ''),
                        'label_style':FieldArg("cssstyle", ''),
                        'action_json':FieldArg("url", ''),
                        'action':FieldArg("url", ''),
                        'hidden_field1':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field2':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field3':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field4':FieldArg("text", '', valdt=True, jsonset=True),
                        'session_storage':FieldArg("text", "", valdt=True, jsonset=True),
                        'local_storage':FieldArg("text","", valdt=True, jsonset=True),
                        'target':FieldArg("text",''),
                        'button_text':FieldArg("text",'Submit'),
                        'button_class':FieldArg("cssclass", ''),
                        'inputdiv_class':FieldArg("cssclass", ''),
                        'inputandbutton_class':FieldArg("cssclass", ''),
                        'inputandbutton_style':FieldArg("cssstyle", ''),
                        'error_class':FieldArg("cssclass", ''),
                        'input_text':FieldArg("text", '', valdt=True, jsonset=True),
                        'size':FieldArg("text", ''),
                        'maxlength':FieldArg("text", ''),
                        'required':FieldArg("boolean", False),
                        'type':FieldArg("text", 'text'),
                        'pattern':FieldArg("text", ''),
                        'title':FieldArg("text", ''),
                        'input_accepted_class':FieldArg("cssclass", ''),
                        'input_errored_class':FieldArg("cssclass", ''),
                        'input_class':FieldArg("cssclass", ''),
                        'set_input_accepted':FieldArg("boolean", False, jsonset=True),
                        'set_input_errored':FieldArg("boolean", False, jsonset=True),
                        'hide':FieldArg("boolean", False, jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        label: The text displayed to the left of the text input field
        label_class: The css class of the label
        action_json:  if a value set, and client has jscript enabled, this is the page ident, label, url this button links to, expects a json page back
        action: The page ident, label, url this button links to, overridden if action_json is set.
        hidden_field1: A hidden field value, leave blank if unused, name used as the get field name
        hidden_field2: A second hidden field value, leave blank if unused, name used as the get field name
        hidden_field3: A third hidden field value, leave blank if unused, name used as the get field name
        hidden_field4: A fourth hidden field value, leave blank if unused, name used as the get field name
        session_storage: A session storage key, this widgfield returns the stored value if anything
        local_storage: A local storage key, this widgfield returns the stored value if anything
        target: if given, the target attribute will be set
        button_text: The text on the button
        button_class: The class given to the button tag
        inputdiv_class: the class attribute of the div which contains the label, input text and button
        inputandbutton_class: the class attribute of the span which contains the input text and button
        error_class: The class applied to the paragraph containing the error message on error.
        input_text: The default text in the text input field, field name used as the name attribute
        size: The number of characters appearing in the text input area
        maxlength: The maximum number of characters accepted in the text area
        required: Set True to put the 'required' flag into the input field
        type: the type set on the input field, such as text or email
        pattern: regular expression pattern
        title: helps describe the pattern
        input_accepted_class: A class which can be set on the input field
        input_errored_class: A class which can be set on the input field
        input_class: Class set on the input field
        set_input_accepted: If True, input_accepted_class will be set on the input field
        set_errored_accepted: If True, input_errored_class will be set on the input field
        hide: If True, widget is hidden
        """
        Widget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "div"
        # error div at 0
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''
        # The form
        self[1] = tag.Part(tag_name='form', attribs={"role":"form", "method":"post"})

        # div containing label, input text and button
        self[1][0] = tag.Part(tag_name='div')
        # the label
        self[1][0][0] = tag.Part(tag_name="label", hide_if_empty=True)
        # span containing input text and button
        self[1][0][1] = tag.Part(tag_name='span')
        # the text input field
        self[1][0][1][0] = tag.ClosedPart(tag_name="input", attribs ={"type":"text"})
        # the submit button
        self[1][0][1][1] = tag.Part(tag_name="button", attribs ={"type":"submit"})
        self[1][0][1][1][0] = "Submit"


    def _build(self, page, ident_list, environ, call_data, lang):
        "build the form"
        if self.wf.target:
            self[1].attribs["target"] = self.wf.target
        # Hides widget if no error and hide is True
        self.widget_hide(self.wf.hide)

        if self.wf.error_class:
            self[0].attribs["class"] = self.wf.error_class
        if self.error_status:
            del self[0].attribs["style"]
        if not self.wf.action:
            # setting self._error replaces the entire tag
            self._error = "Warning: No form action"
            return
        actionurl = self.get_url(self.wf.action)
        if not actionurl:
            # setting self._error replaces the entire tag
            self._error = "Warning: broken link"
            return

        jsonurl = self.get_url(self.wf.action_json)
        # any label:value added to self.jlabels will be set in a javascript fieldvalues attribute for the widget
        if jsonurl:
            self.jlabels['url'] = jsonurl
        if self.wf.input_accepted_class:
           self.jlabels['input_accepted_class'] = self.wf.input_accepted_class
        if self.wf.input_errored_class:
           self.jlabels['input_errored_class'] = self.wf.input_errored_class
        if self.wf.session_storage:
           self.jlabels['session_storage'] = self.wf.session_storage
        if self.wf.local_storage:
           self.jlabels['local_storage'] = self.wf.local_storage

        # update the action of the form
        self[1].attribs["action"] = actionurl
        # the div holding label, input text and button
        if self.wf.inputdiv_class:
            self[1][0].attribs["class"] = self.wf.inputdiv_class

        self[1][0][0].set_class_style(self.wf.label_class, self.wf.label_style)
        if self.wf.label:
            self[1][0][0][0] = self.wf.label

        # the span holding input text and button
        self[1][0][1].set_class_style(self.wf.inputandbutton_class, self.wf.inputandbutton_style)

        # set an id in the input field for the 'label for' tag
        self.jlabels['input_id'] = self[1][0][1][0].insert_id()
        # set the label 'for' attribute
        self[1][0][0].attribs['for'] = self.jlabels['input_id']

        self[1][0][1][0].attribs["name"] = self.get_formname('input_text')
        self[1][0][1][0].attribs["value"] = self.wf.input_text

        if self.wf.size:
            self[1][0][1][0].attribs["size"] = self.wf.size
        if self.wf.maxlength:
            self[1][0][1][0].attribs["maxlength"] = self.wf.maxlength
        if self.wf.required:
            self[1][0][1][0].attribs["required"] = "required"
        if self.wf.type:
            self[1][0][1][0].attribs["type"] = self.wf.type
        if self.wf.pattern:
            self[1][0][1][0].attribs["pattern"] = self.wf.pattern
        if self.wf.title:
            self[1][0][1][0].attribs["title"] = self.wf.title


        if self.wf.input_class:
            input_class = self.wf.input_class
        else:
            input_class = ''

        if self.error_status and self.wf.input_errored_class:
            if input_class:
                input_class = input_class + ' ' + self.wf.input_errored_class
            else:
                input_class = self.wf.input_errored_class
        elif self.wf.set_input_errored and self.wf.input_errored_class:
            if input_class:
                input_class = input_class + ' ' + self.wf.input_errored_class
            else:
                input_class =self.wf.input_errored_class
        elif self.wf.set_input_accepted and self.wf.input_accepted_class:
            if input_class:
                input_class = input_class + ' ' + self.wf.input_accepted_class
            else:
                input_class = self.wf.input_accepted_class

        if input_class:
            self[1][0][1][0].attribs["class"] = input_class

        # submit button
        if self.wf.button_class:
            self[1][0][1][1].attribs["class"] = self.wf.button_class
        if self.wf.button_text:
            self[1][0][1][1][0] = self.wf.button_text

        # add ident and four hidden fields
        self.add_hiddens(self[1], page)



    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets a submit event handler"""
        ident=self.get_id()
        return f"""  $("#{ident} form").on("submit input", function(e) {{
    SKIPOLE.widgets["{ident}"].eventfunc(e);
    }});
"""


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div> <!-- with widget id and class widget_class -->
  <div>  <!-- div hidden when no error is displayed, with class set to error_class on error -->
    <p> <!-- error message appears in this paragraph --> </p>
  </div>
  <form role="form" method="post"> <!-- action attribute set to action field -->
    <div> <!-- class attribute set to inputdiv_class -->
      <label> <!-- with class set to label_class and content to label, for set to input text id -->
      </label>
      <span>  <!-- class attribute set to inputandbutton_class -->
          <input type="text" />  <!-- class set to input_class -->
            <!-- input text value set to input_text -->
          <button type="submit"> <!-- with class set to button_class -->
            <!-- button_text -->
          </button>
      </span>
    </div>
    <!-- hidden input fields -->
  </form>
</div>"""



class SubmitTextInput4(Widget):
    """Defines a form with a text input field, and four hidden fields"""

    # js_validators is a class attribute, True if javascript validation is enabled
    js_validators=True

    error_location = (0,0,0)

    arg_descriptions = {'left_text':FieldArg("text", 'Your input:'),
                        'left_class':FieldArg("cssclass", ''),
                        'left_style':FieldArg("cssstyle", ''),
                        'leftdiv_class':FieldArg("cssclass", ''),
                        'leftdiv_style':FieldArg("cssstyle", ''),
                        'action_json':FieldArg("url", ''),
                        'action':FieldArg("url", ''),
                        'error_class':FieldArg("cssclass", ''),
                        'hidden_field1':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field2':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field3':FieldArg("text", '', valdt=True, jsonset=True),
                        'hidden_field4':FieldArg("text", '', valdt=True, jsonset=True),
                        'target':FieldArg("text",''),
                        'button_text':FieldArg("text",'Submit'),
                        'button_class':FieldArg("cssclass", ''),
                        'formdiv_class':FieldArg("cssclass", ''),
                        'formdiv_style':FieldArg("cssstyle", ''),
                        'inputdiv_class':FieldArg("cssclass", ''),
                        'inputdiv_style':FieldArg("cssstyle", ''),
                        'buttondiv_class':FieldArg("cssclass", ''),
                        'buttondiv_style':FieldArg("cssstyle", ''),
                        'input_text':FieldArg("text", '', valdt=True, jsonset=True),
                        'size':FieldArg("text", ''),
                        'maxlength':FieldArg("text", ''),
                        'required':FieldArg("boolean", False),
                        'type':FieldArg("text", 'text'),
                        'pattern':FieldArg("text", ''),
                        'title':FieldArg("text", ''),
                        'input_accepted_class':FieldArg("cssclass", ''),
                        'input_errored_class':FieldArg("cssclass", ''),
                        'input_class':FieldArg("cssclass", ''),
                        'input_style':FieldArg("cssstyle", ''),
                        'set_input_accepted':FieldArg("boolean", False, jsonset=True),
                        'set_input_errored':FieldArg("boolean", False, jsonset=True),
                        'hide':FieldArg("boolean", False, jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        left_text: The text displayed to the left of the text input field
        leftdiv_class: The css class of the div containing the left paragraph
        leftdiv_style: The css style of the div containing the left paragraph
        left_class: The css class of the p tag containing left_text
        left_style: The css style of the p tag containing left_text
        action_json:  if a value set, and client has jscript enabled, this is the page ident, label, url this button links to, expects a json page back
        action: The page ident, label, url this button links to, overridden if action_json is set.
        error_class: The class applied to the paragraph containing the error message on error.
        hidden_field1: A hidden field value, leave blank if unused, name used as the get field name
        hidden_field2: A second hidden field value, leave blank if unused, name used as the get field name
        hidden_field3: A third hidden field value, leave blank if unused, name used as the get field name
        hidden_field4: A fourth hidden field value, leave blank if unused, name used as the get field name
        target: if given, the target attribute will be set
        button_text: The text on the button
        button_class: The class given to the button tag
        formdiv_class: the class attribute of the div which contains the left text, input field and button
        formdiv_style: the style attribute of the div which contains the left text, input field and button
        inputdiv_class: the class attribute of the div which contains the input field
        inputdiv_style: the style attribute of the div which contains the input field
        buttondiv_class: the class attribute of the div which contains the submit button
        buttondiv_style: the style attribute of the div which contains the submit button
        input_text: The default text in the text input field, field name used as the name attribute
        size: The number of characters appearing in the text input area
        maxlength: The maximum number of characters accepted in the text area
        required: Set True to put the 'required' flag into the input field
        type: the type set on the input field, such as text or email
        pattern: regular expression pattern
        title: helps describe the pattern
        input_accepted_class: A class which can be set on the input field
        input_errored_class: A class which can be set on the input field
        input_class: Class set on the input field
        input_style: Style set on the input field
        set_input_accepted: If True, input_accepted_class will be set on the input field
        set_errored_accepted: If True, input_errored_class will be set on the input field
        hide: If True, widget is hidden
        """
        Widget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "div"
        # error div at 0
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''

        # The form
        self[1] = tag.Part(tag_name='form', attribs={"role":"form", "method":"post"})

        # div containing left text, input text and button
        self[1][0] = tag.Part(tag_name='div')
        # the left div
        self[1][0][0] = tag.Part(tag_name="div", hide_if_empty=True)
        # the left paragraph
        self[1][0][0][0] = tag.Part(tag_name="p", hide_if_empty=True)
        # div containing input field
        self[1][0][1] = tag.Part(tag_name='div')
        # the text input field
        self[1][0][1][0] = tag.ClosedPart(tag_name="input", attribs ={"type":"text"})
        # div containing the submit button
        self[1][0][2] = tag.Part(tag_name='div')
        # the submit button
        self[1][0][2][0] = tag.Part(tag_name="button", attribs ={"type":"submit"})
        self[1][0][2][0][0] = "Submit"


    def _build(self, page, ident_list, environ, call_data, lang):
        "build the form"
        if self.wf.target:
            self[1].attribs["target"] = self.wf.target
        # Hides widget if no error and hide is True
        self.widget_hide(self.wf.hide)

        if self.wf.error_class:
            self[0].attribs["class"] = self.wf.error_class
        if self.error_status:
            del self[0].attribs["style"]
        if not self.wf.action:
            # setting self._error replaces the entire tag
            self._error = "Warning: No form action"
            return
        actionurl = self.get_url(self.wf.action)
        if not actionurl:
            # setting self._error replaces the entire tag
            self._error = "Warning: broken link"
            return

        jsonurl = self.get_url(self.wf.action_json)
        # any label:value added to self.jlabels will be set in a javascript fieldvalues attribute for the widget
        if jsonurl:
            self.jlabels['url'] = jsonurl
        if self.wf.input_accepted_class:
           self.jlabels['input_accepted_class'] = self.wf.input_accepted_class
        if self.wf.input_errored_class:
           self.jlabels['input_errored_class'] = self.wf.input_errored_class


        # update the action of the form
        self[1].attribs["action"] = actionurl
        # the div holding label, input text and button
        self[1][0].set_class_style(self.wf.formdiv_class, self.wf.formdiv_style)

        self[1][0][0].set_class_style(self.wf.leftdiv_class, self.wf.leftdiv_style)

        self[1][0][0][0].set_class_style(self.wf.left_class, self.wf.left_style)
        if self.wf.left_text:
            self[1][0][0][0][0] = self.wf.left_text

        # the div holding the input field
        self[1][0][1].set_class_style(self.wf.inputdiv_class, self.wf.inputdiv_style)

        # set an id in the input field
        self.jlabels['input_id'] = self[1][0][1][0].insert_id()

        self[1][0][1][0].attribs["name"] = self.get_formname('input_text')
        self[1][0][1][0].attribs["value"] = self.wf.input_text

        if self.wf.size:
            self[1][0][1][0].attribs["size"] = self.wf.size
        if self.wf.maxlength:
            self[1][0][1][0].attribs["maxlength"] = self.wf.maxlength
        if self.wf.required:
            self[1][0][1][0].attribs["required"] = "required"
        if self.wf.type:
            self[1][0][1][0].attribs["type"] = self.wf.type
        if self.wf.pattern:
            self[1][0][1][0].attribs["pattern"] = self.wf.pattern
        if self.wf.title:
            self[1][0][1][0].attribs["title"] = self.wf.title
        if self.wf.input_style:
            self[1][0][1][0].attribs["style"] = self.wf.input_style


        if self.wf.input_class:
            input_class = self.wf.input_class
        else:
            input_class = ''

        if self.error_status and self.wf.input_errored_class:
            if input_class:
                input_class = input_class + ' ' + self.wf.input_errored_class
            else:
                input_class = self.wf.input_errored_class
        elif self.wf.set_input_errored and self.wf.input_errored_class:
            if input_class:
                input_class = input_class + ' ' + self.wf.input_errored_class
            else:
                input_class =self.wf.input_errored_class
        elif self.wf.set_input_accepted and self.wf.input_accepted_class:
            if input_class:
                input_class = input_class + ' ' + self.wf.input_accepted_class
            else:
                input_class = self.wf.input_accepted_class

        if input_class:
            self[1][0][1][0].attribs["class"] = input_class

        # the div holding the button
        self[1][0][2].set_class_style(self.wf.buttondiv_class, self.wf.buttondiv_style)
        # submit button
        if self.wf.button_class:
            self[1][0][2][0].attribs["class"] = self.wf.button_class
        if self.wf.button_text:
            self[1][0][2][0][0] = self.wf.button_text
        # add ident and four hidden fields
        self.add_hiddens(self[1], page)


    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets a submit event handler"""
        ident=self.get_id()
        return f"""  $("#{ident} form").on("submit input", function(e) {{
    SKIPOLE.widgets["{ident}"].eventfunc(e);
    }});
"""


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div> <!-- with widget id and class widget_class -->
  <div>  <!-- div hidden when no error is displayed, with class set to error_class on error -->
    <p> <!-- error message appears in this paragraph --> </p>
  </div>
  <form role="form" method="post"> <!-- action attribute set to action field -->
    <div> <!-- class attribute set to formdiv_class -->
      <div> <!-- class set to leftdiv_class -->
          <p>  <!-- class set to left_class -->
              <!-- content set to left_text -->
          </p>
      </div>
      <div>  <!-- class attribute set to inputdiv_class -->
          <input type="text" />  <!-- class set to input_class -->
            <!-- input text value set to input_text -->
      </div>
      <div>  <!-- class attribute set to buttondiv_class -->
          <button type="submit"> <!-- with class set to button_class -->
            <!-- button_text -->
          </button>
      </div>
    </div>
    <!-- hidden input fields -->
  </form>
</div>"""





"""This module defines various widgets which provide information"""
import time

from .. import skiboot
from .. import tag
from . import Widget, ClosedWidget, FieldArg, FieldArgList, FieldArgTable, FieldArgDict


class ServerTimeStamp(Widget):
    """A widget containing the current server time. consists of a span with time stamp"""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {
                        'timestamp':FieldArg("text", '', jsonset=True),
                        'utc':FieldArg("boolean", True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        timestamp: normally empty and timestamp will automatically be displayed, or other text can be set here.
        utc: If True, time is utc, if False it is local time
        """
        Widget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "span"

    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the element"
        # set timer or string
        if self.wf.timestamp:
            self[0] = self.wf.timestamp
        elif self.wf.utc:
            self[0] = time.strftime("%c", time.gmtime())
        else:
            self[0] = time.strftime("%c", time.localtime())

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<span>  <!-- with widget id and class widget_class -->
  <!-- Normally timestamp -->
</span>
"""


class PageIdent(Widget):
    """A widget displaying the ident of a page label
       If no page label or span_text are given,
       shows ident of the current page"""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'page_label':FieldArg("text", ''),
                        'span_text':FieldArg("text", "", jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        page_label: The label of a page or folder
        span_text: if given, overrides the page ident value
        """
        Widget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "span"

    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the element"
        if self.wf.span_text:
            self[0] = self.wf.span_text
        elif self.wf.page_label:
            proj = skiboot.getproject(self.proj_ident)
            # its a label string of this project
            if proj is None:
                self[0] = f"Unable to resolve the ident of {self.wf.page_label}"
                return
            value = proj.resolve_label(self.wf.page_label)
            # so value is one of the tuple ident, None, or URL
            if not value:
                self[0] = f"Unable to resolve the ident of {self.wf.page_label}"
                return
            if isinstance(value, str):
                # a url
                self[0] = f"Given label points to URL rather than a page ident"
                return
            # an ident tuple
            self[0] = f"{value[0]},{value[1]}"
        else:
            self[0] = page.ident.to_comma_str()


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<span>  <!-- with widget id and class widget_class -->
  <!-- string value of the page ident, or span_text -->
</span>
"""


class PageName(Widget):
    """A widget containing the given page name which is set within the text
         If no page ident is given, shows name of the current page"""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'page_ident':FieldArg("text", ''),
                        'span_text':FieldArg("text", "", jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        page_ident: The label or ident of a page or folder
        span_text: if given, overrides the page name value
        """
        Widget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "span"

    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the element"
        if self.wf.span_text:
            self[0] = self.wf.span_text
        elif self.wf.page_ident:
            proj = skiboot.getproject(self.proj_ident)
            if "/" in self.wf.page_ident:
                # its a URL
                self[0] = f"Unable to resolve {self.wf.page_ident}"
                return

            if self.wf.page_ident.isdigit():
                # Its "integer"
                try:
                    value = (self.proj_ident, int(self.wf.page_ident))
                except:
                    self[0] = f"Unable to resolve the ident of {self.wf.page_ident}"
                    return
            elif ',' in self.wf.page_ident:
                # Its a string such as "proj_ident,number"
                vallist = self.wf.page_ident.split(',')
                if len(vallist) != 2:
                    self[0] = f"Unable to resolve the ident of {self.wf.page_ident}"
                    return
                lblproj = vallist[0].strip(" ()[]\"\'")
                lblval = vallist[1].strip(" ()[]\"\'")
                if lblval.isdigit():
                    # Its "subproject, integer"
                    try:
                        value = (lblproj, int(lblval))
                    except:
                        self[0] = f"Unable to resolve the ident of {self.wf.page_ident}"
                        return
                else:
                    # Its "subproject, label"
                    subproj = skiboot.getproject(lblproj)
                    if subproj is None:
                        self[0] = f"Unable to resolve the ident of {self.wf.page_ident}"
                        return
                    value = subproj.resolve_label(lblval)
            else:
                 # its a label string of this project
                if proj is None:
                    self[0] = f"Unable to resolve the ident of {self.wf.page_ident}"
                    return
                value = proj.resolve_label(self.wf.page_ident)

            # so value is one of the tuple ident, None, or URL
            if not value:
                self[0] = f"Unable to resolve the ident of {self.wf.page_ident}"
                return
            if isinstance(value, str):
                # a url
                self[0] = f"Given label points to URL rather than an ident"
                return
            # an ident tuple, make an ident
            page_ident = skiboot.make_ident(value)
            requested_page = skiboot.get_item(page_ident)
            if requested_page is None:
                self[0] = "Unknown page"
            else:
                self[0] = requested_page.name
        else:
            self[0] = page.name


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<span>  <!-- with widget id and class widget_class -->
  <!-- The page name, or span_text -->
</span>
"""


class PageDescription(Widget):
    """A widget containing the page brief of the given ident which is set within the text
         If no page ident is given, shows brief of the current page"""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'page_ident':FieldArg("text", ''),
                        'span_text':FieldArg("text", "", jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        page_ident: The label or ident of a page or folder
        span_text: if given, overrides the page description value
        """
        Widget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "span"

    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the element"
        if self.wf.span_text:
            self[0] = self.wf.span_text
        elif self.wf.page_ident:
            proj = skiboot.getproject(self.proj_ident)
            if "/" in self.wf.page_ident:
                # its a URL
                self[0] = f"Unable to resolve {self.wf.page_ident}"
                return

            if self.wf.page_ident.isdigit():
                # Its "integer"
                try:
                    value = (self.proj_ident, int(self.wf.page_ident))
                except:
                    self[0] = f"Unable to resolve the ident of {self.wf.page_ident}"
                    return
            elif ',' in self.wf.page_ident:
                # Its a string such as "proj_ident,number"
                vallist = self.wf.page_ident.split(',')
                if len(vallist) != 2:
                    self[0] = f"Unable to resolve the ident of {self.wf.page_ident}"
                    return
                lblproj = vallist[0].strip(" ()[]\"\'")
                lblval = vallist[1].strip(" ()[]\"\'")
                if lblval.isdigit():
                    # Its "subproject, integer"
                    try:
                        value = (lblproj, int(lblval))
                    except:
                        self[0] = f"Unable to resolve the ident of {self.wf.page_ident}"
                        return
                else:
                    # Its "subproject, label"
                    subproj = skiboot.getproject(lblproj)
                    if subproj is None:
                        self[0] = f"Unable to resolve the ident of {self.wf.page_ident}"
                        return
                    value = subproj.resolve_label(lblval)
            else:
                 # its a label string of this project
                if proj is None:
                    self[0] = f"Unable to resolve the ident of {self.wf.page_ident}"
                    return
                value = proj.resolve_label(self.wf.page_ident)

            # so value is one of the tuple ident, None, or URL
            if not value:
                self[0] = f"Unable to resolve the ident of {self.wf.page_ident}"
                return
            if isinstance(value, str):
                # a url
                self[0] = f"Given label points to URL rather than an ident"
                return
            # an ident tuple, make an ident
            page_ident = skiboot.make_ident(value)
            requested_page = skiboot.get_item(page_ident)
            if requested_page is None:
                self[0] = "Unknown page"
            else:
                self[0] = requested_page.brief
        else:
            self[0] = page.brief

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<span>  <!-- with widget id and class widget_class -->
  <!-- The page description -->
</span>
"""


class ProjectName(Widget):
    """A span showing the name of the project the page is in"""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {}

    def __init__(self, name=None, brief='', **field_args):
        """
        Shows project name
        """
        Widget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "span"

    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the element"        
        self[0] = page.ident.proj

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<span>  <!-- with widget id and class widget_class -->
  <!-- the project ident (usually the project name unless set otherwise) -->
</span>
"""


class Version(Widget):
    """A span showing the project version of the page it is in"""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {}

    def __init__(self, name=None, brief='', **field_args):
        """
        Shows project version
        """
        Widget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "span"

    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the element"        
        self[0] = page.project.version

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<span>  <!-- with widget id and class widget_class -->
  <!-- string of the project version -->
</span>
"""


class SkipoleVersion(Widget):
    """A span showing this version of the skipole framework"""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {}

    def __init__(self, name=None, brief='', **field_args):
        """
        Shows skipole version
        """
        Widget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "span"

    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the element"        
        self[0] = skiboot.version()

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<span>  <!-- with widget id and class widget_class -->
  <!-- string of the skipole framework version -->
</span>
"""


class Redirector(Widget):
    """A widget containing javascript which redirects the page to the url
       Once the page is loaded, the redirection occurs, so normally
       this is the only widget on the page
       A textblock is displayed with the link to the url if the client has javascript disabled

       If an error is raised with this widget as target, the error message should be a URL which the widget will
       then use - causing a redirection on error."""

    arg_descriptions = {'url':FieldArg("text", ''),
                        'textblock_ref':FieldArg("textblock_ref", ""),
                        'text_refnotfound':FieldArg("text", "If this page does not redirect automatically, follow this link:"),
                        'text_replaceblock':FieldArg("text", "If this page does not redirect automatically, follow this link:"),
                        'linebreaks':FieldArg("boolean", True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        url: the url to redirect to, either a URL containing / character or a label
        textblock_ref: The reference of the TextBlock appearing in the paragraph
        text_refnotfound: text to appear if the textblock is not found
        text_replaceblock: text set here will replace the textblock
        linebreaks: Set True if linebreaks in the text are to be shown as html breaks
        """
        Widget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "div"
        self[0] = tag.Part(tag_name='script', attribs={"type":"text/javascript"})
        self[1] = tag.Part(tag_name='p')
        self._url = ''


    def _error_build(self, message):
        """Called if an error is raised"""
        # sets the error message as the widget URL
        if message:
            self._url = message


    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the widget"
        if self.error_status and self._url:
            url = self._url
        else:
            url = self.wf.url

        if not url:
            self[1][0] = "Invalid URL given"
            return

        if "/" not in url:
            # not a valid url, could be a label
            proj = skiboot.getproject(self.proj_ident)
            if proj is None:
                self[1][0] = "Unable to resolve the URL"
                return
            value = proj.label_to_url(url)
            if value is None:
                self[1][0] = f"Invalid URL \"{url}\" has been given"
                return
            url = value
        
        self[0][0] = f"window.location.replace(\"{url}\");"

        linebreaks = bool(self.wf.linebreaks)

        if self.wf.text_replaceblock:
            # no textblock, just the replacement text
            tblock = self.wf.text_replaceblock
            if not linebreaks:
                self[1].linebreaks = False
        else:
            # define the textblock
            tblock = self.wf.textblock_ref
            tblock.failmessage = self.wf.text_refnotfound
            tblock.linebreaks = linebreaks
            tblock.proj_ident = page.proj_ident

        self[1][0] = tblock
        self[1][1] = tag.ClosedPart(tag_name='br')
        self[1][2] = tag.Part(tag_name='a', attribs={"href":url}, text=url)


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
  <!-- url is either the url argument, or is the error message if an error is raised on the widget -->
  <script type="text/javascript">
     <!-- window.location.replace("url"); -->
  </script>
  <p>
    <!-- text_replaceblock or the textblock text here -->
    <br />
    <a href="url"> 
        <!-- url --> 
    </a>
  </p>
</div>
"""


class ProgressBar1(Widget):
    """A div containing a progress bar"""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'label':FieldArg("text", "Progress:"),
                        'label_class':FieldArg("cssclass", ''),
                        'label_style':FieldArg("cssstyle", ''),
                        'text':FieldArg("text", "0%", jsonset=True),
                        'value':FieldArg("text", "0", jsonset=True),
                        'max':FieldArg("text", "100"),
                        'progress_class':FieldArg("cssclass", ''),
                        'progress_style':FieldArg("cssstyle", ''),
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        text shown in place of the bar if the client browser does not support the progress tag
        """
        Widget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "div"
        self[0] = tag.Part(tag_name="label", hide_if_empty=True)
        self[1] = tag.Part(tag_name='progress')

    def _build(self, page, ident_list, environ, call_data, lang):
        "build the widget"
        self[0].set_class_style(self.wf.label_class, self.wf.label_style)
        if self.wf.label:
            self[0][0] = self.wf.label
        # set an id in the progress tag for the 'label for' tag
        for_id = self[1].insert_id()
        # any label:value added to self.jlabels will be set in a javascript fieldvalues attribute for the widget
        self.jlabels['progressident'] = for_id
        self[0].attribs['for'] = for_id

        self[1].set_class_style(self.wf.progress_class, self.wf.progress_style)
        if self.wf.text:
            self[1][0] = self.wf.text
        if self.wf.max:
            self[1].attribs["max"] = self.wf.max
        if self.wf.value:
            self[1].attribs["value"] = self.wf.value


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
 <label> <!-- with class label_class and style label_style -->
         <!-- with contents label -->
 </label>
 <progress> <!-- with attributes value and max -->
            <!-- with contents text -->
 </progress> 
</div>
"""



class ProgressBar2(Widget):
    """A div containing a progress bar, uses width of a div rather than the progress tag"""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'value':FieldArg("integer", "0", jsonset=True),
                        'bar_class':FieldArg("cssclass", '')
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        value is an integer representing a percentage between 0 and 100
        bar_class should set the bar colour, such as w3-green
        """
        Widget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "div"
        self[0] = tag.Part(tag_name="div")


    def _build(self, page, ident_list, environ, call_data, lang):
        "build the widget"
        if self.wf.value<0:
            value = 0
        elif self.wf.value>100:
            value = 100
        else:
            value = self.wf.value
        barstyle = f"height:24px;width:{value}%"
        self[0].set_class_style(self.wf.bar_class, barstyle)


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
 <div> <!-- with class bar_class and style width set to value -->
 </div>
</div>
"""






"""A WSGI application generator

skilift is an associated python package used to develop an application.

Typically a developer's PC would have both the skilift and the skipole Python packages installed. Skilift includes a development web server and provides a web admin interface, which together with your own code and the skipole functions, enables the developer to create a WSGI application.

Once created, your application and its support files can be moved to your deployment server, which also needs a WSGI compatible web server, and the skipole package. 

The deployment server does not need the skilift application.

Skipole and skilift require python 3.8 or later, and can be installed with:

python3 -m pip install skipole

python3 -m pip install skilift


skilift
-------

To generate a new project, and use a web admin interface to develop it, the package skilift is required.

skilift can be run from the command line with the python -m option

Usage is

python3 -m skilift mynewproj /path/to/projectfiles

Which creates a directory /path/to/projectfiles
containing sub directory mynewproj - containing project data, and file
mynewproj.py where your code will be developed.

You should replace 'mynewproj' with your preferred name for a new project.

The path "/path/to/projectfiles" is the path to a directory where you will
develop your project. Multiple projects can be created in one 'projectfiles'
directory, or you could have multiple such directories holding different
projects.

If mynewproj already exists in the directory, it will not be changed.

You should then inspect the file

/path/to/projectfiles/mynewproj.py

where your code will be developed.


skipole
-------

skipole is intended to be imported, if it is run using

python3 -m skipole

this text is displayed.

If it is run using

python3 -m skipole --version

Then a version string is displayed.

When imported skipole makes the following available:

version - a version string of the form a.b.c

WSGIApplication - an instance of this class is a callable WSGI application (see below)

set_debug(mode) - a function to turn on debugging (if mode is True), or off (if mode is False)

use_submit_list - is available to optionally wrap the user defined submit_data function
                  Enables a responder 'submit list' to define package,module,function to
                  be called as the responder's submit_data function, where package,module
                  is relative to the users code.

PageData - An instance of this class is used to update page widgets.

SectionData - An instance of this class is used to update section widgets

widget_modules() - Return a tuple of widget module names

widgets_in_module(module_name) - Returns a tuple of widget names present in the module


Exceptions
----------

These are also provided, and can be raised within the users code:

ValidateError - returns the project validation error page

ServerError - returns the project server error page

GoTo - diverts the call to another page

FailPage - diverts the call to the calling Responder's 'Fail page'

ServeFile - sends a static server file to the client browser


WSGIApplication
---------------

An instance of this class should be created, and is a callable WSGI application.

The WSGIApplication has the following arguments which should be provided to create
an instance:

PROJECT - the project name

PROJECTFILES - the directory containing your projects

PROJ_DATA - An optional dictionary you may wish to provide

start_call - a function you should create, called at the start of a call

submit_data - a function you should create, called by responders

end_call - a function you should create, called at the end of the call, prior to returning the page

url - path where this project will be served, typically '/'

proj_ident - project identifier, normally None which auto sets it to the PROJECT value

You would typically define your functions, and then create an instance:

my_application = WSGIApplication(project=PROJECT,
                                 projectfiles=PROJECTFILES,
                                 proj_data=PROJ_DATA,
                                 start_call=start_call,
                                 submit_data=submit_data,
                                 end_call=end_call,
                                 url="/",
                                 proj_ident=None)

If the argument proj_ident is left at its default None value, then it will be automatically
set to the project name. However it can be set to a different string here which may be useful
if multiple instances of this project are to be created and added to a parent 'root' project.
Each unique proj_ident will then define each of the sub applications.

This my_application is then a callable WSGI application.

The WSGIApplication class has method:

add_project(self, proj, url, check_cookies=None) - adds other projects to the 'root' project.

Where proj is another instance of a WSGIApplication and will be served at the path
given by argument url.

The optional check_cookies argument can be set to a function which you would create, with signature:
def my_check_cookies_function(received_cookies, proj_data):
Before the call is routed to the subapplication, your my_check_cookies_function is called, with the
received_cookies dictionary, and with your application's proj_data dictionary. If your function
returns None, the call proceeds unhindered to the subapplication. If however your function returns
an ident tuple, of the form (proj_ident, pagenumber), then the call is routed to that page instead.

The skis module has the function makeapp() which creates a project providing needed javascript
files which should be added to your application, for example:

from skipole import skis
skis_application = skis.makeapp()
my_application.add_project(skis_application, url='/lib')

Which causes the skis project to be served at /lib.

"""

import sys, traceback, inspect, pkgutil

from functools import wraps
from importlib import import_module

from .ski import skiboot, widgets

from .ski.project_class_definition import SkipoleProject, PageData, SectionData
from .ski.excepts import ValidateError, ServerError, GoTo, FailPage, ServeFile

version = skiboot.version()


__all__ = ['WSGIApplication', 'ValidateError', 'ServerError', 'GoTo', 'FailPage', 'ServeFile',
           'set_debug', 'use_submit_list', 'version', 'PageData', 'SectionData', 'widget_modules', 'widgets_in_module']


class WSGIApplication(object):
    """The WSGIApplication - an instance being a callable WSGI application"""

    def __init__(self, project, projectfiles, proj_data={}, start_call=None, submit_data=None, end_call=None, url="/", proj_ident=None):
        """An instance of this class is a callable WSGI application.

Arguments are:

project - the project name

projectfiles - the directory containing your projects

proj_data - an optional dictionary, passed as an attribute of skicall,
            which is an object passed to your functions during the progress of a call

start_call - a function you should create, called at the start of a call

submit_data - a function you should create, called by responders

end_call - a function you should create, called at the end of the call, prior to returning the page

url - path where this project will be served, typically '/'

proj_ident - generally set to None, in which case it will be auto set to the project name
"""
        self._skipoleproject = SkipoleProject(project, projectfiles, proj_data, start_call, submit_data, end_call, url, proj_ident)


    @property
    def project(self):
        """The project name"""
        return self._skipoleproject.proj_name

    @property
    def proj_ident(self):
        """The project ident"""
        return self._skipoleproject.proj_ident

    @property
    def projectfiles(self):
        """The projectfiles path"""
        return self._skipoleproject.projectfiles

    @property
    def proj_data(self):
        """The proj_data dictionary"""
        return self._skipoleproject.proj_data

    @property
    def start_call(self):
        """The start_call function"""
        return self._skipoleproject.start_call

    @property
    def submit_data(self):
        """The submit_data function"""
        return self._skipoleproject.submit_data

    @property
    def end_call(self):
        """The end_call function"""
        return self._skipoleproject.end_call

    @property
    def url(self):
        """This project path"""
        return self._skipoleproject.url

    @property
    def brief(self):
        """This project brief description"""
        return self._skipoleproject.brief

    @property
    def version(self):
        """This project version string"""
        return self._skipoleproject.version

    def __call__(self, environ, start_response):
        "The instance is callable"
        return self._skipoleproject(environ, start_response)

    def add_project(self, proj, url=None, check_cookies=None):
        """Add a sub project to this root project, returns the sub project url
           proj is the sub project WSGIApplication object."""
        return self._skipoleproject.add_project(proj._skipoleproject, url, check_cookies)

    def set_accesstextblocks(self, accesstextblocks):
        """Set an instance of a class which reads and writes TextBlocks. The default class is defined in the skipole.textblocks module,
           which simply stores TextBlocks in memory after reading them from a JSON file, and is not suitable for the dynamic creation
           of TextBlocks. It is sufficient for the single user of Skiadmin, but is essentially read only when serving your pages.
           If required, your own object could be created which should support the methods of the default class, and could read and write
           from your own database. This opens the possibility of dynamic creation and alteration of TextBlocks."""
        self._skipoleproject.textblocks = accesstextblocks


def set_debug(mode):
    "If mode is True, this sets increased debug error messages to be displayed"
    skiboot.set_debug(mode)



# This 'use_submit_list' is available to wrap the submit_data function if required

def use_submit_list(submit_data):
    "Used to decorate submit_data to enable submit_list to define package,module,function"
    @wraps(submit_data)
    def submit_function(skicall):
        "This function replaces submit_data, if skicall.submit_list has two or more elements"
        if not skicall.submit_list:
            # do nothing, simply return the original submit_data
            return submit_data(skicall)
        if len(skicall.submit_list) < 2:
            # do nothing, simply return the original submit_data
            return submit_data(skicall)
        # get the module where submit_data is defined
        sdmodule = inspect.getmodule(submit_data)
        if sdmodule.__name__ == "__main__":
            # absolute path
            submitpath = ".".join(skicall.submit_list[:-1])
            sdpackage = None
        else:
            # relative path
            submitpath = "." + ".".join(skicall.submit_list[:-1])
            sdpackage = sdmodule.__name__
        try:
            submitmodule = import_module(submitpath, sdpackage)
        except Exception as e:
            raise ServerError("Unable to import project module defined as %s in submit list" % (submitpath,)) from e
        # now obtain and run the specified function
        try:
            submitfunc = getattr(submitmodule, skicall.submit_list[-1])
        except Exception as e:
            raise ServerError("submit_list package %s found, but the required function %s is not recognised" % (submitpath, skicall.submit_list[-1])) from e
        return submitfunc(skicall)
    return submit_function


def widget_modules():
    "Return a tuple of widget module names"
    return tuple(name for (module_loader, name, ispkg) in pkgutil.iter_modules(widgets.__path__))


def widgets_in_module(module_name):
    "Returns a tuple of widget names present in the module"
    module = import_module('.ski.widgets.' + module_name, __name__)
    widget_list = []
    for classname,obj in inspect.getmembers(module, lambda member: inspect.isclass(member) and (member.__module__ == module.__name__)):
        widget_list.append(classname)
    return tuple(widget_list)







"""Contains widgets for lists"""

from .. import tag
from . import Widget, ClosedWidget, FieldArg, FieldArgList, FieldArgTable, FieldArgDict, AnchorClickEventMixin


class UList1(Widget):
    """An unordered list widget."""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {
                       'highlight_class':FieldArg("cssclass", ''),
                       'set_highlight':FieldArg("boolean", False, jsonset=True),
                       'contents': FieldArgList('text', jsonset=True),
                       'even_class':FieldArg("cssclass", ""),
                       'odd_class':FieldArg("cssclass", "")
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        highlight_class: CSS class will replace widget_class if set_highlight is True
        set_highlight: Set to True to force highlight_class to replace widget_class
                       Set to False to remove highlight_class and return to widget_class
        contents: List of text strings to be shown, gives the number of list items
        even_class: class of even li elements, if empty string, then no class will be applied
        odd_class: class of odd li elements, if empty string, then no class will be applied
        """
        Widget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "ul"


    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the list"
        contents = self.wf.contents

        # any label:value added to self.jlabels will be set in a javascript fieldvalues attribute for the widget
        if self.wf.highlight_class:
            self.jlabels['highlight_class'] = self.wf.highlight_class
            if self.wf.set_highlight:
                self.attribs["class"] = self.wf.highlight_class

        # set even li class
        if self.wf.even_class:
            _even = self.wf.even_class
            self.jlabels['even_class'] = _even
        else:
            _even = ''
        # set odd li class
        if self.wf.odd_class:
            _odd = self.wf.odd_class
            self.jlabels['odd_class'] = _odd
        else:
            _odd = ''
        for index, item in enumerate(contents):
            if _even and (index % 2):
                self[index] = tag.Part(tag_name="li", attribs={"class":_even}, text=item)
            elif _odd and not (index % 2):
                self[index] = tag.Part(tag_name='li', attribs={"class":_odd}, text=item)
            else:
                self[index] = tag.Part(tag_name='li', text=item)


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """<ul>  <!-- with widget id and class widget_class -->
  <li> <!-- with style from odd_class or even_class -->
      <!-- with text from each contents element -->
  </li>
  <!-- with appropriate number of list elements -->
</ul>
"""


class UList2(Widget):
    """An unordered list widget containing html"""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {
                       'highlight_class':FieldArg("cssclass", ''),
                       'set_highlight':FieldArg("boolean", False, jsonset=True),
                       'set_html': FieldArgList('text', jsonset=True),
                       'even_class':FieldArg("cssclass", ""),
                       'odd_class':FieldArg("cssclass", "")
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        highlight_class: CSS class will replace widget_class if set_highlight is True
        set_highlight: Set to True to force highlight_class to replace widget_class
                       Set to False to remove highlight_class and return to widget_class
        set_html: List of html strings to be shown, html will be shown unescaped
        even_class: class of even li elements, if empty string, then no class will be applied
        odd_class: class of odd li elements, if empty string, then no class will be applied
        """
        Widget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "ul"

    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the list"
        contents = self.wf.set_html

        # any label:value added to self.jlabels will be set in a javascript fieldvalues attribute for the widget

        if self.wf.highlight_class:
            self.jlabels['highlight_class'] = self.wf.highlight_class
            if self.wf.set_highlight:
                self.attribs["class"] = self.wf.highlight_class

        # set even li class
        if self.wf.even_class:
            _even = self.wf.even_class
            self.jlabels['even_class'] = _even
        else:
            _even = ''
        # set odd li class
        if self.wf.odd_class:
            _odd = self.wf.odd_class
            self.jlabels['odd_class'] = _odd
        else:
            _odd = ''

        for index, item in enumerate(contents):
            if _even and (index % 2):
                self[index] = tag.Part(tag_name="li", attribs={"class":_even}, text=item)
            elif _odd and not (index % 2):
                self[index] = tag.Part(tag_name='li', attribs={"class":_odd}, text=item)
            else:
                self[index] = tag.Part(tag_name='li', text=item)
            self[index].htmlescaped = False
            self[index].linebreaks=False


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """<ul>  <!-- with widget id and class widget_class -->
  <li> <!-- with style from odd_class or even_class -->
      <!-- with html from each set_html element -->
  </li>
  <!-- with appropriate number of list elements -->
</ul>
"""


class TableList(AnchorClickEventMixin, Widget):
    """A table of a single text column, followed by Up, Down and Remove button link columns
       Each link having an individual get field set
       Typically used to display and order a Python list"""


    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {
                        'highlight_class':FieldArg("cssclass", ''),
                        'set_highlight':FieldArg("boolean", False, jsonset=True),
                        'even_class':FieldArg("cssclass", ""),
                        'odd_class':FieldArg("cssclass", ""),
                        'maximize_text_col':FieldArg('boolean',True),
                        'hide':FieldArg("boolean", False, jsonset=True),
                        'button_class':FieldArg("cssclass", ""),
                        'remove_button_text':FieldArg('text', 'Remove'),
                        'up_link_ident':FieldArg("url", 'no_javascript'),
                        'up_json_ident':FieldArg("url", ''),
                        'down_link_ident':FieldArg("url", 'no_javascript'),
                        'down_json_ident':FieldArg("url", ''),
                        'remove_link_ident':FieldArg("url", 'no_javascript'),
                        'remove_json_ident':FieldArg("url", ''),
                        'contents':FieldArgTable(['text', 'text', 'text', 'text'], valdt=True, jsonset=True)
                        }

    def __init__(self, name=None, brief='', **field_args):
        """
        highlight_class: CSS class will replace widget_class if set_highlight is True
        set_highlight: Set to True to force highlight_class to replace widget_class
                       Set to False to remove highlight_class and return to widget_class
        even_class: class of even rows, if empty string, then no class will be applied
        odd_class: class of odd rows, if empty string, then no class will be applied
        maximize_text_col: If True the text column is made large, with the button columns reduced to the size of the buttons
        hide: if True, table will be hidden
        button_class: The CSS class to apply to the buttons
        remove_button_text: Text on the Remove button
        up_link_ident: For the up arrow button: the target page link ident, label or url, if javascript not available
        up_json_ident: For the up arrow button: the url, ident or label to link, expecting a json file to be returned
        down_link_ident: For the down arrow button: the target page link ident, label or url, if javascript not available
        down_json_ident: For the down arrow button: the url, ident or label to link, expecting a json file to be returned
        remove_link_ident: For the Remove button: the target page link ident, label or url, if javascript not available
        remove_json_ident: For the Remove button: the url, ident or label to link, expecting a json file to be returned
        contents: col 0 is the text string to place in the first column,
                  col 1 is the get field contents of the up button link
                  col 2 is the get field contents of the down button link
                  col 3 is the get field contents of the remove button link
                  This fieldname used as the widgfield
        """
        Widget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "table"


    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the table"
        # Hides widget if no error and hide is True
        self.widget_hide(self.wf.hide)
        fieldtable = self.wf.contents

        # any label:value added to self.jlabels will be set in a javascript fieldvalues attribute for the widget

        if self.wf.highlight_class:
            self.jlabels['highlight_class'] = self.wf.highlight_class
            if self.wf.set_highlight:
                self.attribs["class"] = self.wf.highlight_class
        self.jlabels['button_class'] = self.wf.button_class
        self.jlabels['remove_button_text'] = self.wf.remove_button_text

        # set even row colour
        if self.wf.even_class:
            even = self.wf.even_class
        else:
            even = ''
        self.jlabels['even_class'] = even

        # set odd row colour
        if self.wf.odd_class:
            odd = self.wf.odd_class
        else:
            odd = ''
        self.jlabels['odd_class'] = odd

        if self.wf.up_json_ident:
            self.jlabels['up_json_url'] = self.get_url(self.wf.up_json_ident)
        if self.wf.down_json_ident:
            self.jlabels['down_json_url'] = self.get_url(self.wf.down_json_ident)
        if self.wf.remove_json_ident:
            self.jlabels['remove_json_url'] = self.get_url(self.wf.remove_json_ident)

        if self.wf.up_link_ident:
            up_url = self.get_url(self.wf.up_link_ident)
            self.jlabels['up_link_url'] = up_url
        else:
            up_url = ''

        if self.wf.down_link_ident:
            down_url = self.get_url(self.wf.down_link_ident)
            self.jlabels['down_link_url'] = down_url
        else:
            down_url = ''

        if self.wf.remove_link_ident:
            remove_url = self.get_url(self.wf.remove_link_ident)
            self.jlabels['remove_link_url'] = remove_url
        else:
            remove_url = ''

        if self.wf.maximize_text_col:
           self.jlabels['maximize_text_col'] = True

        if not fieldtable:
            return
        len_fieldtable = len(fieldtable)
        lastrow = len_fieldtable - 1

        # create rows
        for rownumber, row in enumerate(fieldtable):
            if even and (rownumber % 2) :
                self[rownumber] = tag.Part(tag_name="tr", attribs={"class":even})
            elif odd and not (rownumber % 2):
                self[rownumber] = tag.Part(tag_name='tr', attribs={"class":odd})
            else:
                self[rownumber] = tag.Part(tag_name='tr')

            # first column is text
            if self.wf.maximize_text_col:
                self[rownumber][0] = tag.Part(tag_name='td', text = row[0], attribs={"style":"width : 100%;"})
            else:
                self[rownumber][0] = tag.Part(tag_name='td', text = row[0])

            # Next column is up button link
            self[rownumber][1] = tag.Part(tag_name='td')
            if not rownumber:
                # rownumber 0 has no up button
                self[rownumber][1][0] = tag.HTMLSymbol("&nbsp;")
            else:
                if self.wf.button_class:
                    self[rownumber][1][0] = tag.Part(tag_name='a', attribs = {"role":"button", "class":self.wf.button_class})
                else:
                    self[rownumber][1][0] = tag.Part(tag_name='a', attribs = {"role":"button"})
                self[rownumber][1][0].htmlescaped=False
                if up_url:
                    self[rownumber][1][0][0] = tag.HTMLSymbol("&uarr;")
                    # create a url for the href
                    get_fields = {self.get_formname("contents"):row[1]}
                    self[rownumber][1][0].attribs["href"] = self.make_get_url(page, up_url, get_fields, True)
                else:
                    self[rownumber][1][0] = "Warning: broken link"

            # Next column is down button link
            self[rownumber][2] = tag.Part(tag_name='td')
            if rownumber == lastrow:
                # last row has no down button
                self[rownumber][2][0] = tag.HTMLSymbol("&nbsp;")
            else:
                if self.wf.button_class:
                    self[rownumber][2][0] = tag.Part(tag_name='a', attribs = {"role":"button", "class":self.wf.button_class})
                else:
                    self[rownumber][2][0] = tag.Part(tag_name='a', attribs = {"role":"button"})
                self[rownumber][2][0].htmlescaped=False
                if down_url:
                    self[rownumber][2][0][0] = tag.HTMLSymbol("&darr;")
                    # create a url for the href
                    get_fields = {self.get_formname("contents"):row[2]}
                    self[rownumber][2][0].attribs["href"] = self.make_get_url(page, down_url, get_fields, True)
                else:
                    self[rownumber][2][0] = "Warning: broken link"

            # Next column is Remove button link
            self[rownumber][3] = tag.Part(tag_name='td')
            if self.wf.button_class:
                self[rownumber][3][0] = tag.Part(tag_name='a', attribs = {"role":"button", "class":self.wf.button_class})
            else:
                self[rownumber][3][0] = tag.Part(tag_name='a', attribs = {"role":"button"})
            self[rownumber][3][0].htmlescaped=False
            if remove_url:
                if self.wf.remove_button_text:
                    self[rownumber][3][0][0] = self.wf.remove_button_text
                else:
                    self[rownumber][3][0][0] = "Remove"
                # create a url for the href
                get_fields = {self.get_formname("contents"):row[3]}
                self[rownumber][3][0].attribs["href"] = self.make_get_url(page, remove_url, get_fields, True)
            else:
                self[rownumber][3][0] = "Warning: broken link"



    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<table>  <!-- with widget id and class widget_class -->
                   <!-- and attribute style=display:none if hide is True -->
  <tr> <!-- with class from even or odd classes -->
    <td> <!-- text string --> </td>
    <td>
      <a role="button" href="#">
      <!-- With class set by button_class, and the href link will be the url of the up_link_ident with get field -->
        &uarr;
      </a>
    </td>
    <td>
      <a role="button" href="#">
      <!-- With class set by button_class, and the href link will be the url of the down_link_ident with get field -->
        &darr;
      </a>
    </td>
    <td>
      <a role="button" href="#">
      <!-- With class set by button_class, and the href link will be the url of the remove_link_ident with get field -->
        Remove
      </a>
    </td>
  </tr>
  <!-- rows repeated -->
</table>"""


class FileList(Widget):
    """Defines a div, containing links to downloadable files
       If the download parameter is true, the file pointed by the url is downloaded rather than displayed in the browser"""

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {
                        'file_links':FieldArgTable(("url", "text", "boolean"), valdt=True),    # target, filename, download
                        'button_class':FieldArg("cssclass", ''),
                        'button_style':FieldArg("cssstyle", '')
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
         file_links: A list of lists, each inner list describing a link
        For each link, the table row is
          0 : The url, label or ident of the target page of the link
          1 : Filename - The displayed text of the link and name of file to be downloaded
          2 : True if the link is to be downloaded rather than displayed
        button_class: The class of the button links - which provides the appearance via CSS
        """
        Widget.__init__(self, name=name, brief=brief, **field_args)
        self.tag_name = "div"
        # as this widget does not display errors, and is not json settable, hide if empty
        self.hide_if_empty=True


    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the list of links"
        for row in self.wf.file_links:
            linkurl, filename, download = row
            if not (linkurl or filename):
                continue
            url = self.get_url(linkurl)
            if not url:
                continue
            if download:
                lnk = tag.Part(tag_name="a", text=filename, attribs={"href":url, "role":"button", "download":filename})
            else:
                lnk = tag.Part(tag_name="a", text=filename, attribs={"href":url, "role":"button"})
            lnk.set_class_style(self.wf.button_class, self.wf.button_style)
            self.append(lnk)

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>  <!-- with widget id and class widget_class -->
  <a role=\"button\" href=\"#\" download=\"filename\">  <!-- with class set to button_class -->
    <!-- The filename -->
  </a>
  <!-- further links -->
</div>"""






# If called using 
#
# python3 -m skipole
#
#

import sys

from . import version


if len(sys.argv) == 2:
    # if called with "python3 -m skipole --version"
    if sys.argv[1] == "--version":
        print(version)
        sys.exit(0)

print(f"""skipole version {version}

skipole is a WSGI application generator.

skilift is an associated python package used to develop an application.

Typically a developer's PC would have both the skilift and the skipole Python packages installed. Skilift includes a development web server and provides a web admin interface, which together with your own code and the skipole functions, enables the developer to create a WSGI application.

Once created, your application and its support files can be moved to your deployment server, which also needs a WSGI compatible web server, and the skipole package. 

The deployment server does not need the skilift application.

Skipole and skilift require python 3.8 or later, and can be installed with:

python3 -m pip install skipole

python3 -m pip install skilift


skilift
-------

To generate a new project, and use a web admin interface to develop it, the package skilift is required.

skilift can be run from the command line with the python -m option

Usage is

python3 -m skilift mynewproj /path/to/projectfiles

Which creates a directory /path/to/projectfiles
containing sub directory mynewproj - containing project data, and file
mynewproj.py where your code will be developed.

You should replace 'mynewproj' with your preferred name for a new project.

The path "/path/to/projectfiles" is the path to a directory where you will
develop your project. Multiple projects can be created in one 'projectfiles'
directory, or you could have multiple such directories holding different
projects.

If mynewproj already exists in the directory, it will not be changed.

You should then inspect the file

/path/to/projectfiles/mynewproj.py

where your code will be developed.


skipole
-------

skipole is intended to be imported, if it is run using

python3 -m skipole

this text is displayed.

If it is run using

python3 -m skipole --version

Then a version string is displayed.

When imported skipole makes the following available:

version - a version string of the form a.b.c

WSGIApplication - an instance of this class is a callable WSGI application (see below)

set_debug(mode) - a function to turn on debugging (if mode is True), or off (if mode is False)

use_submit_list - is available to optionally wrap the user defined submit_data function
                  Enables a responder 'submit list' to define package,module,function to
                  be called as the responder's submit_data function, where package,module
                  is relative to the users code.

PageData - An instance of this class is used to update page widgets.

SectionData - An instance of this class is used to update section widgets

widget_modules() - Return a tuple of widget module names

widgets_in_module(module_name) - Returns a tuple of widget names present in the module


Exceptions
----------

These are also provided, and can be raised within the users code:

ValidateError - returns the project validation error page

ServerError - returns the project server error page

GoTo - diverts the call to another page

FailPage - diverts the call to the calling Responder's 'Fail page'

ServeFile - sends a static server file to the client browser


WSGIApplication
---------------

An instance of this class should be created, and is a callable WSGI application.

The WSGIApplication has the following arguments which should be provided to create
an instance:

PROJECT - the project name

PROJECTFILES - the directory containing your projects

PROJ_DATA - An optional dictionary you may wish to provide

start_call - a function you should create, called at the start of a call

submit_data - a function you should create, called by responders

end_call - a function you should create, called at the end of the call, prior to returning the page

url - path where this project will be served, typically '/'

proj_ident - project identifier, normally None which auto sets it to the PROJECT value

You would typically define your functions, and then create an instance:

my_application = WSGIApplication(project=PROJECT,
                                 projectfiles=PROJECTFILES,
                                 proj_data=PROJ_DATA,
                                 start_call=start_call,
                                 submit_data=submit_data,
                                 end_call=end_call,
                                 url="/",
                                 proj_ident=None)

If the argument proj_ident is left at its default None value, then it will be automatically
set to the project name. However it can be set to a different string here which may be useful
if multiple instances of this project are to be created and added to a parent 'root' project.
Each unique proj_ident will then define each of the sub applications.

This my_application is then a callable WSGI application.

The WSGIApplication class has method:

add_project(self, proj, url, check_cookies=None) - adds other projects to the 'root' project.

Where proj is another instance of a WSGIApplication and will be served at the path
given by argument url.

The optional check_cookies argument can be set to a function which you would create, with signature:
def my_check_cookies_function(received_cookies, proj_data):
Before the call is routed to the subapplication, your my_check_cookies_function is called, with the
received_cookies dictionary, and with your application's proj_data dictionary. If your function
returns None, the call proceeds unhindered to the subapplication. If however your function returns
an ident tuple, of the form (proj_ident, pagenumber), then the call is routed to that page instead.

The skis module has the function makeapp() which creates a project providing needed javascript
files which should be added to your application, for example:

from skipole import skis
skis_application = skis.makeapp()
my_application.add_project(skis_application, url='/lib')

Which causes the skis project to be served at /lib.

""")


